import numpy as np
from matplotlib import cm
from porteratzolibs.visualization_o3d.open3d_utils import (
    convertcloud,
)
import copy

try:
    import open3d

    V3V = open3d.utility.Vector3dVector
    use_headless = False
except ImportError:
    use_headless = True


def check_point_format(nppoints):
    assert (
        (type(nppoints) == np.ndarray)
        or (type(nppoints) is list)
        or (type(nppoints) is tuple)
        or (type(nppoints) is o3d_pointSetClass)
    ), "Not valid point_cloud"

    if (type(nppoints) is not list) & (type(nppoints) is not tuple):
        if type(nppoints) is np.ndarray:
            if len(nppoints.shape) == 2:
                nppoints = [nppoints]
        if type(nppoints) is o3d_pointSetClass:
            nppoints = [nppoints]
    return nppoints


class pointSetClass:
    def __init__(self, points=None, color=[], persistant=False) -> None:
        self.colors = None
        self.points = None
        self.persistant = persistant
        self.updated = False
        if points is not None:
            self.update(points, color)

    def update(self, points, color=[]):
        self.points = points
        try:
            if len(np.asarray(color)) == len(self.points):
                if len(np.asarray(color).shape) == 2:
                    self.colors = color
                else:
                    color = color - np.min(color)
                    self.colors = cm.jet(color / np.max(color))[:, :3]
            elif len(color) > 0:
                self.colors = np.ones_like(points) * color
            self.updated = True
        except TypeError:
            pass
        except ValueError:
            pass

    def draw(self):
        pass

    def state_dict(self):
        if (self.persistant) or (self.updated):
            self.updated = False
            return {
                "points": self.points.astype(np.float32)
                if type(self.points) is np.ndarray
                else self.points,
                "colors": self.colors.astype(np.float32)
                if type(self.colors) is np.ndarray
                else self.colors,
            }
        else:
            return {}


class o3d_pointSetClass(pointSetClass):
    def __init__(self, points=None, color=[], persistant=False) -> None:
        self.cloud = None
        self.in_vis = False
        self.used_vis = None
        super().__init__(points, color, persistant)

    def update(self, points=[], color=[]):
        if len(points) > 0:
            super().update(points, color)
        if self.cloud is None:
            self.cloud = convertcloud(self.points)
        else:
            self.cloud.points = V3V(self.points)
        if self.colors is not None:
            self.cloud.colors = V3V(self.colors)
        self.updated = True

    def draw(self, vis):
        if (self.persistant) or (self.updated):
            if not self.in_vis:
                vis.add_geometry(self.cloud)
                self.in_vis = True
            else:
                vis.update_geometry(self.cloud)
            self.updated = False
        else:
            self.remove(vis)

    def remove(self, vis):
        vis.remove_geometry(self.cloud)
        self.in_vis = False

class multi_bounding_box:
    def __init__(self, workpoints, color) -> None:
        self.main_bb = open3d.geometry.OrientedBoundingBox.create_from_points(
            open3d.utility.Vector3dVector(workpoints)
        )
        self.main_bb.color = color
        self.scale_factor = 0.05
        self.number_of_lines = 8

        self.secondary_bb = []
        for i in range(self.number_of_lines):
            self.secondary_bb.append(copy.copy(self.main_bb))
            self.secondary_bb[-1].scale(
                1.0 + self.scale_factor + i * self.scale_factor,
                self.secondary_bb[-1].get_center(),
            )

    def set_points(self, workpoints):
        main_bb = open3d.geometry.OrientedBoundingBox.create_from_points(
            open3d.utility.Vector3dVector(workpoints)
        )
        self.main_bb.center = main_bb.center
        self.main_bb.extent = main_bb.extent
        self.main_bb.R = main_bb.R

        for n, i in enumerate(self.secondary_bb):
            sec = copy.copy(main_bb)
            sec.scale(1.0 + self.scale_factor + n * self.scale_factor, sec.get_center())
            i.center = sec.center
            i.extent = sec.extent
            i.R = sec.R

    def draw(self, vis):
        vis.add_geometry(self.main_bb)
        for i in self.secondary_bb:
            vis.add_geometry(i)

    def update(self, vis):
        vis.update_geometry(self.main_bb)
        for i in self.secondary_bb:
            vis.update_geometry(i)

    def remove(self, vis):
        vis.remove_geometry(self.main_bb)
        for i in self.secondary_bb:
            vis.remove_geometry(i)


import numpy as np
import random
import math
import os
import sys
import pclpy
import open3d as o3d
from porteratzolibs.ThreeDProcessing import pclpy_utils 
from TreeToolML.utils.tictoc import bench_dict


def downsample(point_cloud, leaf_size=0.005, return_idx=False):
    if return_idx:
        pcd = o3d.geometry.PointCloud()
        pcd.points = o3d.utility.Vector3dVector(point_cloud)
        _, rest, _ = pcd.voxel_down_sample_and_trace(
            voxel_size=leaf_size,
            min_bound=np.array([-10, -10, -10]),
            max_bound=np.array([10, 10, 10]),
        )
        return rest[rest != -1]
    else:
        return pclpy_utils.voxelize(point_cloud, leaf_size)


def outliers(points, min_n=6, radius=0.4, organized=True):
    _points = pclpy_utils.radius_outlier_removal(points, min_n, radius, organized)
    return _points[~np.all(np.isnan(_points), axis=1)]


def combine_IOU(vis_dict):
    vis_dict_ = {n: i for n, i in enumerate(vis_dict)}
    discard_list = []
    ious = []
    for key1, _points in list(vis_dict_.items()):
        if key1 in discard_list:
            continue

        for key2, _points2 in list(vis_dict_.items()):
            if key1 == key2:
                continue

            if key2 in discard_list:
                continue

            min_xyz1 = np.min(_points, axis=0)
            max_xyz1 = np.max(_points, axis=0)

            min_xyz2 = np.min(_points2, axis=0)
            max_xyz2 = np.max(_points2, axis=0)

            box1 = [min_xyz1[0], min_xyz1[1], max_xyz1[0], max_xyz1[1]]
            box2 = [min_xyz2[0], min_xyz2[1], max_xyz2[0], max_xyz2[1]]
            iou = bb_intersection_over_union(box1, box2)
            ious.append(iou)
            if iou > 0.2:
                vis_dict_[key1] = np.vstack([vis_dict_[key1], vis_dict_[key2]])
                vis_dict_.pop(key2)
                discard_list.append(key2)
    return list(vis_dict_.values())


def data_preprocess(temp_point_set):
    temp_xyz = temp_point_set[:, :3]
    temp_xyz = normalize(temp_xyz)
    return temp_xyz


def load_data(path):
    try:
        return np.load(path, allow_pickle=True)
    except:
        return np.loadtxt(path)


def get_data_set(data_path):
    files_set = os.listdir(data_path)
    random.shuffle(files_set)
    return files_set


def get_train_val_set(trainingdata_path, val_rate=0.20):
    train_set = []
    val_set = []
    all_train_set = os.listdir(trainingdata_path)
    random.shuffle(all_train_set)
    total_num = len(all_train_set)
    val_num = int(val_rate * total_num)
    for j in range(len(all_train_set)):
        if j < val_num:
            val_set.append(all_train_set[j])
        else:
            train_set.append(all_train_set[j])
    return train_set, val_set


def normalize_2(sample_xyz):
    centerd_tree = sample_xyz - np.multiply(np.min(sample_xyz, 0), [0, 0, 1])
    centerd_tree = centerd_tree - np.multiply(np.mean(centerd_tree, axis=0), [1, 1, 0])
    # normalize into unit sphere
    centerd_tree /= np.max(np.linalg.norm(centerd_tree, axis=1))
    return centerd_tree


def normalize(sample_xyz):
    min_xyz = np.min(sample_xyz, axis=0)
    max_xyz = np.max(sample_xyz, axis=0)
    deta_central_xyz = (max_xyz - min_xyz) / 2.0
    central_xyz = deta_central_xyz + min_xyz
    n_data = sample_xyz - central_xyz
    # normalize into unit sphere
    n_data /= np.max(np.linalg.norm(n_data, axis=1))
    return n_data


def compute_object_center(sample_xyz):
    min_xyz = np.min(sample_xyz, axis=0)
    max_xyz = np.max(sample_xyz, axis=0)
    deta_central_xyz = (max_xyz - min_xyz) / 2.0
    central_xyz = deta_central_xyz + min_xyz
    return central_xyz


def jitter_point_cloud(sample_xyz, Jitter_argument, sigma=0.001, clip=0.05):
    if np.random.random() < Jitter_argument:
        N, C = sample_xyz.shape
        assert clip > 0
        jittered_data = np.clip(sigma * np.random.randn(N, C), -1 * clip, clip)
        sample_xyz += jittered_data
    return sample_xyz


def shuffle_data(data):
    idx = np.arange(np.size(data, 0))
    np.random.shuffle(idx)
    return data[idx, ...]


def ratation(sample_xyz, Rotation_argument):
    if np.random.random() < Rotation_argument:
        ###
        rot = random.uniform(0, 2 * math.pi)
        rotation_matrix = [
            [math.cos(rot), math.sin(rot), 0],
            [-math.sin(rot), math.cos(rot), 0],
            [0, 0, 1],
        ]
        sample_xyz = np.dot(sample_xyz, rotation_matrix)
    return sample_xyz


def ratation_angle(sample_xyz, angel):
    rot = angel / 180.0
    rotation_matrix = [
        [math.cos(rot), math.sin(rot), 0],
        [-math.sin(rot), math.cos(rot), 0],
        [0, 0, 1],
    ]
    sample_xyz = np.dot(sample_xyz, rotation_matrix)
    return sample_xyz


def transfer_xy(sample_xyz, x_d, y_d):
    temp_ones = np.ones([np.size(sample_xyz, 0), 1])
    sample_xyz = np.concatenate([sample_xyz, temp_ones], axis=-1)

    transfer_matrix = [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 1], [x_d, y_d, 0, 1]]
    sample_xyz = np.dot(sample_xyz, transfer_matrix)
    return sample_xyz[:, :3]


def farthest_point_sample(xyz, npoint):
    N, _ = xyz.shape
    centroids = []
    distance = np.ones(N) * 1e10
    farthest = np.random.randint(0, N)
    for i in range(npoint):
        centroids.append(farthest)
        centroid = xyz[farthest, :]
        dist = np.sum((xyz - centroid) ** 2, -1)
        mask = dist < distance
        distance[mask] = dist[mask]
        farthest = int(np.where(distance == np.max(distance))[0][0])
    return centroids


def makesphere(centroid=[0, 0, 0], radius=1, dense=90):
    n = np.arange(0, 360, int(360 / dense))
    n = np.deg2rad(n)
    x, y = np.meshgrid(n, n)
    x = x.flatten()
    y = y.flatten()
    sphere = np.vstack(
        [
            centroid[0] + np.sin(x) * np.cos(y) * radius,
            centroid[1] + np.sin(x) * np.sin(y) * radius,
            centroid[2] + np.cos(x) * radius,
        ]
    ).T
    return sphere


def bb_intersection_over_union(boxA, boxB):
    # determine the (x, y)-coordinates of the intersection rectangle
    xA = max(boxA[0], boxB[0])
    yA = max(boxA[1], boxB[1])
    xB = min(boxA[2], boxB[2])
    yB = min(boxA[3], boxB[3])
    interArea = max(0, xB - xA + 1) * max(0, yB - yA + 1)
    boxAArea = (boxA[2] - boxA[0] + 1) * (boxA[3] - boxA[1] + 1)
    boxBArea = (boxB[2] - boxB[0] + 1) * (boxB[3] - boxB[1] + 1)
    iou = interArea / float(boxAArea + boxBArea - interArea)
    return iou


def get_center_scale(sample_xyz):
    min_xyz = np.min(sample_xyz, axis=0)
    max_xyz = np.max(sample_xyz, axis=0)
    deta_central_xyz = (max_xyz - min_xyz) / 2.0
    central_xyz = deta_central_xyz + min_xyz
    n_data = sample_xyz - central_xyz
    # normalize into unit sphere
    scale = np.max(np.linalg.norm(n_data, axis=1))
    return central_xyz, scale


def normal_filter(
        subject_cloud,
        search_radius=0.05,
        verticality_threshold=0.3,
        curvature_threshold=0.3,
        return_indexes=False,
):
    bench_dict["normals"].gstep()
    non_ground_normals = pclpy_utils.extract_normals(subject_cloud, search_radius)
    bench_dict["normals"].step("extract")
    # remove Nan points
    non_nan_mask = np.bitwise_not(np.isnan(non_ground_normals.normals[:, 0]))
    non_nan_cloud = subject_cloud[non_nan_mask]
    non_nan_normals = non_ground_normals.normals[non_nan_mask]
    non_nan_curvature = non_ground_normals.curvature[non_nan_mask]
    bench_dict["normals"].step("nan")

    # get mask by filtering verticality and curvature
    verticality = np.dot(non_nan_normals, [[0], [0], [1]])
    verticality_mask = (verticality < verticality_threshold) & (
            -verticality_threshold < verticality
    )
    curvature_mask = non_nan_curvature < curvature_threshold
    verticality_curvature_mask = verticality_mask.ravel() & curvature_mask.ravel()

    only_horizontal_points = non_nan_cloud[verticality_curvature_mask]
    bench_dict["normals"].step("filter")
    bench_dict["normals"].gstop()

    if return_indexes:
        out_index = non_nan_mask
        out_index[non_nan_mask] = verticality_curvature_mask
        return out_index
    else:
        return only_horizontal_points


def seg_normals(
        filtered_points,
        search_radius=0.05,
        normalweight=0.01,
        miter=1000,
        distance=0.01,
        rlim=[0, 0.2],
):
    indices, model = pclpy_utils.segment_normals(
        filtered_points,
        search_radius=search_radius,
        model=pclpy.pcl.sample_consensus.SACMODEL_CYLINDER,
        method=pclpy.pcl.sample_consensus.SAC_RANSAC,
        normalweight=normalweight,
        miter=miter,
        distance=distance,
        rlim=rlim,
    )
    return indices, model


def trunk_center(filtered_points):
    # half_height = (np.max(filtered_points[:, 2]) - np.min(filtered_points[:, 2])) / 2
    # filtered_points = filtered_points[filtered_points[:, 2] < half_height]
    indices, _ = seg_normals(filtered_points)
    temp_object_center_xyz = np.mean(filtered_points[indices], 0)
    if len(filtered_points[indices]) < len(filtered_points) * 0.1:
        return []
    return temp_object_center_xyz


def group_trees(
        filtered_points, tolerance=0.1, min_cluster_size=20, max_cluster_size=25000
):
    cluster_list = pclpy_utils.euclidean_cluster_extract(
        filtered_points,
        tolerance=tolerance,
        min_cluster_size=min_cluster_size,
        max_cluster_size=max_cluster_size,
    )
    return cluster_list


def get_tree_center(
        temp_xyz,
        downsample_leaf=0.01,
        search_radius=0.1,
        verticality_threshold=0.4,
        curvature_threshold=0.1,
        return_filtered=False,
        non_if_no_seg_center=False,
):
    bench_dict["get_centers"].gstep()
    down_points = downsample(temp_xyz, downsample_leaf, False)
    bench_dict["get_centers"].step("start_downsample")
    no_outlier_tree = outliers(down_points, 100, 10)
    bench_dict["get_centers"].step("outliers")
    filtered_points = normal_filter(
        no_outlier_tree,
        search_radius,
        verticality_threshold,
        curvature_threshold,
        return_indexes=False,
    )
    bench_dict["get_centers"].step("filter")
    if len(filtered_points) > len(down_points) * 0.1:
        temp_object_center_xyz = trunk_center(filtered_points)
        if len(temp_object_center_xyz) == 0:
            if non_if_no_seg_center:
                temp_object_center_xyz = None
            else:
                temp_object_center_xyz = np.mean(filtered_points, 0)
    else:
        if non_if_no_seg_center:
            temp_object_center_xyz = None
        else:
            temp_object_center_xyz = np.mean(filtered_points, 0)
    bench_dict["get_centers"].step("trunk")
    bench_dict["get_centers"].gstop()

    if return_filtered:
        return (
            temp_object_center_xyz,
            filtered_points,
        )
    return temp_object_center_xyz



import numpy as np
from matplotlib import cm
from porteratzolibs.visualization_o3d import open3d_utils
import os
import time
from porteratzolibs.visualization_o3d.open3d_pointsetClass import check_point_format, o3d_pointSetClass, pointSetClass, multi_bounding_box
from porteratzolibs.Misc.transforms import eulerAnglesToRotationMatrix

try:
    import open3d
    V3V = open3d.utility.Vector3dVector
    use_headless = False
except ImportError:
    use_headless = True
import pickle


def sidexsidepaint(*point_clusters, color_map="jet", pointsize=0.1, axis=False, for_thesis=False):
    new_points = []
    trans_list = {}
    for arg in point_clusters:
        trans = 0
        for n, points in enumerate(arg):
            if len(points) == 0:
                continue
            _points = points + np.array([[trans, 0, 0]])
            if trans_list.get(n, None) is None:
                trans_list[n] = np.max(_points, 0)[0]
            trans = trans_list[n]
            new_points.append(_points)
    open3dpaint(new_points, pointsize=pointsize, axis=axis, for_thesis=for_thesis)


def open3dpaint(nppoints, color_map="jet", pointsize=0.1, axis=False, for_thesis=False, voxel_size=None):
    nppoints = check_point_format(nppoints)
    if len(nppoints) == 0:
        return
    try:
        if not use_headless:
            vis = open3d.visualization.Visualizer()
            vis.create_window()
            opt = vis.get_render_option()
            if for_thesis:
                opt.background_color = np.asarray([1.0, 1.0, 1.0])
            else:
                opt.background_color = np.asarray([0.1, 0.1, 0.1])
                
            opt.point_size = pointsize
            current_pointset = o3d_pointSetClass()
        else:
            current_pointset = pointSetClass()

        if len(nppoints) > 1:
            if len(nppoints) < 3:
                color_list = [cm.jet(i)[:3] for i in np.linspace(0, 1, len(nppoints))]
            else:
                color_list = [cm.nipy_spectral(i)[:3] for i in np.linspace(0, 1, len(nppoints))]
            choice_list = np.random.choice(len(nppoints), len(nppoints), replace=False)
            for n, i in enumerate(nppoints):
                if type(i) != o3d_pointSetClass:
                    workpoints = i
                    color = color_list[choice_list[n]]
                    if voxel_size is not None:
                        idx = open3d_utils.downsample(workpoints, leaf_size=voxel_size, return_idx=True)
                        workpoints = workpoints[idx]
                    current_pointset = o3d_pointSetClass()
                    current_pointset.update(workpoints, color)
                    current_pointset.draw(vis)
                elif type(i) == o3d_pointSetClass:
                    i.in_vis = False
                    i.update()
                    i.draw(vis)

        else:
            if type(nppoints[0]) != o3d_pointSetClass:
                workpoints = nppoints[0]
                if voxel_size is not None:
                    idx = open3d_utils.downsample(workpoints, leaf_size=voxel_size, return_idx=True)
                    workpoints = workpoints[idx]
                current_pointset.update(workpoints)
                current_pointset.draw(vis)
            elif type(nppoints[0]) == o3d_pointSetClass:
                nppoints[0].in_vis = False
                nppoints[0].update()
                nppoints[0].draw(vis)
        if axis:
            vis.add_geometry(open3d.geometry.TriangleMesh.create_coordinate_frame(axis))
        vis.run()
        vis.clear_geometries()
        vis.destroy_window()

    except Exception as e:
        print(type(e))
        print(e.args)
        print(e)
        vis.destroy_window()


class open3dpaint_non_block:
    def __init__(
        self,
        pointsize=0.1,
        file_name=None,
        axis=False,
        disable=False,
        headless=False,
    ) -> None:
        self.disable = disable
        if use_headless:
            self.headless = True
        else:
            self.headless = headless

        if not self.disable:
            if not self.headless:
                self.vis = open3d.visualization.Visualizer()
                self.vis.create_window(width=640, height=480)
                self.opt = self.vis.get_render_option()
                self.opt.background_color = np.asarray([0.1, 0.1, 0.1])
                self.opt.point_size = pointsize

                self.view = self.vis.get_view_control()
                self.view.set_constant_z_far(10000)
                if axis:
                    self.vis.add_geometry(
                        open3d.geometry.TriangleMesh.create_coordinate_frame(100)
                    )
            self.T = np.eye(4)
            self.all_bb = []
            self.frame = 0
            self.file_name = file_name
            self.pointsets = {}
            self.state_save = []

    def update_points(self, nppoints, pointset=0, color_map="jet", persistant=False):
        if not self.disable:

            if len(nppoints) < 1:
                return
            nppoints = check_point_format(nppoints)

            if len(nppoints) > 1:
                group_points = []
                group_colors = []
                for n, i in enumerate(nppoints):
                    workpoints = i
                    group_points.append(workpoints)
                    colNORM = n / len(nppoints) / 2 + n % 2 * 0.5

                    if color_map == "jet":
                        color = cm.jet(colNORM)[:3]
                        group_colors.append(np.ones_like(workpoints) * color)
                    else:
                        group_colors.append(np.ones_like(workpoints) * color_map)

                if not self.pointsets.get(pointset):
                    if not self.headless:
                        self.pointsets[pointset] = o3d_pointSetClass(
                            persistant=persistant
                        )
                    else:
                        self.pointsets[pointset] = pointSetClass(persistant=persistant)
                self.pointsets[pointset].update(
                    np.concatenate(group_points), np.concatenate(group_colors)
                )

            elif len(nppoints) == 0:
                pass
            else:
                workpoints = nppoints[0]
                if not self.pointsets.get(pointset):
                    if not self.headless:
                        self.pointsets[pointset] = o3d_pointSetClass(
                            persistant=persistant
                        )
                    else:
                        self.pointsets[pointset] = pointSetClass(persistant=persistant)

                self.pointsets[pointset].update(workpoints, color_map)

    def rotate_pointset(self, pointset=0, rotation=(0.0, 0.0, 0.0), center="self"):
        R = self.pointsets[pointset].get_rotation_matrix_from_xyz(np.deg2rad(rotation))
        if type(center) is str:
            self.pointsets[pointset].rotate(R)
        else:
            self.pointsets[pointset].rotate(R, center=center)

    def update_points_of_interest_multiline(
        self, nppoints, color=np.array([1.0, 0.0, 0.0])
    ):
        if not self.disable:
            nppoints = check_point_format(nppoints)
            if not self.headless:
                if len(nppoints) > 1:
                    for n, i in enumerate(nppoints):
                        workpoints = i
                        if len(self.all_bb) < n + 1:
                            if len(workpoints) > 3:
                                bb = multi_bounding_box(workpoints, color)
                                bb.draw(self.vis)
                                # self.vis.add_geometry(bb)
                                self.all_bb.append(bb)
                        else:
                            if len(workpoints) > 3:
                                self.all_bb[n].set_points(workpoints)
                                self.all_bb[n].update(self.vis)
                                # self.vis.update_geometry(self.all_bb[n])
                elif len(nppoints) == 0:
                    pass
                else:
                    workpoints = nppoints[0]

                    if len(self.all_bb) == 0:
                        if len(workpoints) > 3:
                            bb = multi_bounding_box(workpoints, color)
                            bb.draw(self.vis)
                            # self.vis.add_geometry(bb)
                            self.all_bb.append(bb)
                    else:
                        if len(workpoints) > 3:
                            self.all_bb[0].set_points(workpoints)
                            self.all_bb[0].update(self.vis)

                while len(nppoints) < len(self.all_bb):
                    # self.vis.remove_geometry(self.all_bb[-1])
                    self.all_bb[-1].remove(self.vis)
                    self.all_bb.pop(-1)

    def rotate(self, x, y, z):
        if not self.disable:
            T = np.eye(4)
            T[:3, :3] = eulerAnglesToRotationMatrix(np.deg2rad([x, y, z]))
            self.T = T @ self.T

    def translate(self, x, y, z):
        if not self.disable:
            T = np.eye(4)
            T[:, 3] = np.array([x, y, z, 1])
            self.T = T @ self.T

    def set_perspective(self):
        if not self.disable:
            if not self.headless:
                self.cam = self.view.convert_to_pinhole_camera_parameters()
                self.cam.extrinsic = self.T
                self.view.convert_from_pinhole_camera_parameters(self.cam)
                time.sleep(0.02)

    def draw(self, saveim=False, save=None, inc=False):
        if not self.disable:
            for keys in self.pointsets.keys():
                if not self.headless:
                    self.pointsets[keys].draw(self.vis)

            if not self.headless:
                self.vis.poll_events()
                self.vis.update_renderer()

                if saveim:
                    if self.file_name is not None:
                        if save is None:
                            save_path = (
                                self.file_name + str(self.frame).zfill(5) + ".jpg"
                            )
                        else:
                            save_path = save
                            if inc:
                                n = 0
                                while os.path.isfile(save_path):
                                    save_path = (
                                        save.split(".jpg")[0] + "_" + str(n) + ".jpg"
                                    )
                                    n += 1
                        self.vis.capture_screen_image(save_path)
                        self.frame += 1
                return not self.vis.poll_events()
            else:
                if saveim:
                    self.state_save.append(
                        {
                            key: self.pointsets[key].state_dict()
                            for key in self.pointsets.keys()
                        }
                    )
                return False

    def save_state(self):
        with open(self.file_name + ".pkl", "wb") as f:
            pickle.dump(self.state_save, f, pickle.HIGHEST_PROTOCOL)

    def load_state(self, load_path):
        with open(load_path, "rb") as f:
            self.state_save = pickle.load(f)

    def run_from_state(self, save):
        state = self.state_save
        exit_vis = False
        for step in state:
            for pointset_key in step.keys():
                self.update_points(
                    step[pointset_key]["points"],
                    pointset=pointset_key,
                    color_map=step[pointset_key]["colors"],
                )
            self.set_perspective()
            if self.draw(save):
                exit_vis = True
                break
        return exit_vis

    def stop(self):
        if not self.disable:
            if not self.headless:
                self.vis.destroy_window()



import numpy as np
from porteratzolibs.geometry import rotation_matrix_from_vectors

def make_plane(coefs, center=[0, 0, 0]):
    ll, ul = -1, 1
    step = (ul - ll) / 10
    planepoints = np.meshgrid(np.arange(ll, ul, step), np.arange(ll, ul, step))
    plane = np.array(
        [
            planepoints[0].flatten(),
            planepoints[1].flatten(),
            np.zeros(len(planepoints[0].flatten())),
        ]
    ).T
    R = rotation_matrix_from_vectors([0, 0, 1], coefs[0:3])
    return np.add(np.matmul(R, plane.T).T, center)

def make_pointvector(coefs, centroid=[0, 0, 0], length=1, dense=10):
    assert len(coefs) == 3, "Need x,y,z normalvector"
    newcoefs = np.array(coefs) / np.linalg.norm(np.array(coefs))
    pointline = np.arange(
        length / dense, length + length / dense, length / dense
    )
    pointline = np.vstack([pointline, pointline, pointline])
    out = np.add(np.multiply(pointline.T, newcoefs), centroid)
    return out

def make_sphere(centroid=[0, 0, 0], radius=1, dense=90):
    n = np.arange(0, 360, int(360 / dense))
    n = np.deg2rad(n)
    x, y = np.meshgrid(n, n)
    x = x.flatten()
    y = y.flatten()
    sphere = np.vstack(
        [
            centroid[0] + np.sin(x) * np.cos(y) * radius,
            centroid[1] + np.sin(x) * np.sin(y) * radius,
            centroid[2] + np.cos(x) * radius,
        ]
    ).T
    return sphere

def make_cylinder(model=[0, 0, 0, 1, 0, 0, 1], length=1, dense=10):
    radius = model[6]
    X, Y, Z = model[:3]
    direction = model[3:6] / np.linalg.norm(model[3:6])
    n = np.arange(0, 360, int(360 / dense))
    height = np.arange(0, length, length / dense)
    n = np.deg2rad(n)
    x, z = np.meshgrid(n, height)
    x = x.flatten()
    z = z.flatten()
    cyl = np.vstack([np.cos(x) * radius, np.sin(x) * radius, z]).T
    rotation = rotation_matrix_from_vectors([0, 0, 1], model[3:6])
    rotatedcyl = np.matmul(rotation, cyl.T).T + np.array([X, Y, Z])
    return rotatedcyl


import numpy as np
import math


def SVDRigidBodyTransform(Points1, points2):
    centroid1 = np.sum(Points1, 0) / len(Points1)
    centroid2 = np.sum(points2, 0) / len(points2)
    CenteredVector1 = (Points1 - centroid1).transpose()
    CenteredVector2 = (points2 - centroid2).transpose()

    YiT = CenteredVector2.transpose()
    Xi = CenteredVector1
    Wi = np.eye(Xi.shape[1])

    S = np.matmul(Xi, np.matmul(Wi, YiT))
    U, SigNDiag, VT = np.linalg.svd(S, full_matrices=True)
    Raro = np.eye(3)
    Raro[-1, -1] = np.linalg.det(np.matmul(VT.transpose(), U.transpose()))

    RotacionFinal = np.matmul(VT.transpose(), np.matmul(Raro, U.transpose()))
    traslation = (centroid2 - np.matmul(RotacionFinal, centroid1)).reshape(
        3, 1
    )
    return traslation, RotacionFinal

def SVDLstSqr(X, Y):
    U, s, VT = np.linalg.svd(X)
    S = np.diag(s)
    Si = np.zeros([U.shape[1], VT.shape[0]])
    Si[0 : S.shape[0], 0 : S.shape[0]] = np.diag(1 / s)
    return np.matmul(VT.T, np.matmul(Si.T, np.matmul(U.T, Y)))

def eulerAnglesToRotationMatrix(theta):
    R_x = np.array(
        [
            [1, 0, 0],
            [0, np.cos(theta[0]), -np.sin(theta[0])],
            [0, np.sin(theta[0]), np.cos(theta[0])],
        ]
    )
    R_y = np.array(
        [
            [np.cos(theta[1]), 0, np.sin(theta[1])],
            [0, 1, 0],
            [-np.sin(theta[1]), 0, np.cos(theta[1])],
        ]
    )
    R_z = np.array(
        [
            [np.cos(theta[2]), -np.sin(theta[2]), 0],
            [np.sin(theta[2]), np.cos(theta[2]), 0],
            [0, 0, 1],
        ]
    )
    R = np.dot(R_z, np.dot(R_y, R_x))
    return R


def eulerAnglesToRotationMatrixZYX(theta):
    R_x = np.array(
        [
            [1, 0, 0],
            [0, np.cos(theta[0]), -np.sin(theta[0])],
            [0, np.sin(theta[0]), np.cos(theta[0])],
        ]
    )
    R_y = np.array(
        [
            [np.cos(theta[1]), 0, np.sin(theta[1])],
            [0, 1, 0],
            [-np.sin(theta[1]), 0, np.cos(theta[1])],
        ]
    )
    R_z = np.array(
        [
            [np.cos(theta[2]), -np.sin(theta[2]), 0],
            [np.sin(theta[2]), np.cos(theta[2]), 0],
            [0, 0, 1],
        ]
    )
    R = np.dot(R_x, np.dot(R_y, R_z))
    return R


def isRotationMatrix(R):
    Rt = np.transpose(R)
    shouldBeIdentity = np.dot(Rt, R)
    I = np.identity(3, dtype=R.dtype)
    n = np.linalg.norm(I - shouldBeIdentity)
    return n < 1e-6


def rotationMatrixToEulerAnglesZYX(R):
    assert isRotationMatrix(R)

    if R[2, 0] < 1:
        if R[2, 0] > -1:
            y = np.arcsin(-R[2, 0])
            z = np.arctan2(R[1, 0], R[0, 0])
            x = np.arctan2(R[2, 1], R[1, 1])
        else:
            y = np.pi / 2
            z = -np.arctan2(-R[1, 2], R[1, 1])
            x = 0
    else:
        y = -np.pi / 2
        z = np.arctan2(-R[1, 2], R[1, 1])
        x = 0

    return np.array([x, y, z])


def rotationMatrixToEulerAnglesYZX(R):
    assert isRotationMatrix(R)

    if R[1, 0] < 1:
        if R[1, 0] > -1:
            z = np.arcsin(R[1, 0])
            y = np.arctan2(-R[2, 0], R[0, 0])
            x = np.arctan2(-R[1, 2], R[1, 1])
        else:
            z = -np.pi / 2
            y = -np.arctan2(-R[2, 1], R[2, 2])
            x = 0
    else:
        z = -np.pi / 2
        y = np.arctan2(R[2, 1], R[2, 2])
        x = 0

    return np.array([x, y, z])


def rotationMatrixToEulerAngles(R):
    assert isRotationMatrix(R)

    sy = math.sqrt(R[0, 0] * R[0, 0] + R[1, 0] * R[1, 0])

    singular = sy < 1e-6

    if not singular:
        x = math.atan2(R[2, 1], R[2, 2])
        y = math.atan2(-R[2, 0], sy)
        z = math.atan2(R[1, 0], R[0, 0])
    else:
        x = math.atan2(-R[1, 2], R[1, 1])
        y = math.atan2(-R[2, 0], sy)
        z = 0

    return np.array([x, y, z])


import cv2
from matplotlib import cm
import numpy as np
import pandas as pd
import open3d
import math
from pdb import set_trace as Debug


def get_LIDAR(file, names=None, infer=True):
    if infer:
        raw_dataset = pd.read_csv(
            file, na_values="?", comment="\t", sep=",", skipinitialspace=True
        )
    else:
        if names is not None:
            raw_dataset = pd.read_csv(
                file,
                na_values="?",
                comment="\t",
                sep=",",
                skipinitialspace=True,
                names=names,
            )
        else:
            raw_dataset = pd.read_csv(
                file,
                na_values="?",
                comment="\t",
                sep=",",
                skipinitialspace=True,
                header=None,
            )
    return raw_dataset


def convertcloud(points):
    pcd = open3d.geometry.PointCloud()
    pcd.points = open3d.utility.Vector3dVector(points)
    # open3d.write_point_cloud(Path+'sync.ply', pcd)
    # pcd_load = open3d.read_point_cloud(Path+'sync.ply')
    return pcd


def cloudsave(points, Path):
    pcd = open3d.PointCloud()
    pcd.points = open3d.utility.Vector3dVector(points)
    open3d.write_point_cloud(Path + ".ply", pcd)
    pcd_load = open3d.read_point_cloud(Path + ".ply")
    return pcd_load


def eulerAnglesToRotationMatrix(theta):
    R_x = np.array(
        [
            [1, 0, 0],
            [0, np.cos(theta[0]), -np.sin(theta[0])],
            [0, np.sin(theta[0]), np.cos(theta[0])],
        ]
    )
    R_y = np.array(
        [
            [np.cos(theta[1]), 0, np.sin(theta[1])],
            [0, 1, 0],
            [-np.sin(theta[1]), 0, np.cos(theta[1])],
        ]
    )
    R_z = np.array(
        [
            [np.cos(theta[2]), -np.sin(theta[2]), 0],
            [np.sin(theta[2]), np.cos(theta[2]), 0],
            [0, 0, 1],
        ]
    )
    R = np.dot(R_z, np.dot(R_y, R_x))
    return R


def eulerAnglesToRotationMatrixZYX(theta):
    R_x = np.array(
        [
            [1, 0, 0],
            [0, np.cos(theta[0]), -np.sin(theta[0])],
            [0, np.sin(theta[0]), np.cos(theta[0])],
        ]
    )
    R_y = np.array(
        [
            [np.cos(theta[1]), 0, np.sin(theta[1])],
            [0, 1, 0],
            [-np.sin(theta[1]), 0, np.cos(theta[1])],
        ]
    )
    R_z = np.array(
        [
            [np.cos(theta[2]), -np.sin(theta[2]), 0],
            [np.sin(theta[2]), np.cos(theta[2]), 0],
            [0, 0, 1],
        ]
    )
    R = np.dot(R_x, np.dot(R_y, R_z))
    return R


def isRotationMatrix(R):
    Rt = np.transpose(R)
    shouldBeIdentity = np.dot(Rt, R)
    I = np.identity(3, dtype=R.dtype)
    n = np.linalg.norm(I - shouldBeIdentity)
    return n < 1e-6


def rotationMatrixToEulerAnglesZYX(R):
    assert isRotationMatrix(R)

    if R[2, 0] < 1:
        if R[2, 0] > -1:
            y = np.arcsin(-R[2, 0])
            z = np.arctan2(R[1, 0], R[0, 0])
            x = np.arctan2(R[2, 1], R[1, 1])
        else:
            y = np.pi / 2
            z = -np.arctan2(-R[1, 2], R[1, 1])
            x = 0
    else:
        y = -np.pi / 2
        z = np.arctan2(-R[1, 2], R[1, 1])
        x = 0

    return np.array([x, y, z])


def rotationMatrixToEulerAnglesYZX(R):
    assert isRotationMatrix(R)

    if R[1, 0] < 1:
        if R[1, 0] > -1:
            z = np.arcsin(R[1, 0])
            y = np.arctan2(-R[2, 0], R[0, 0])
            x = np.arctan2(-R[1, 2], R[1, 1])
        else:
            z = -np.pi / 2
            y = -np.arctan2(-R[2, 1], R[2, 2])
            x = 0
    else:
        z = -np.pi / 2
        y = np.arctan2(R[2, 1], R[2, 2])
        x = 0

    return np.array([x, y, z])


def rotationMatrixToEulerAngles(R):
    assert isRotationMatrix(R)

    sy = math.sqrt(R[0, 0] * R[0, 0] + R[1, 0] * R[1, 0])

    singular = sy < 1e-6

    if not singular:
        x = math.atan2(R[2, 1], R[2, 2])
        y = math.atan2(-R[2, 0], sy)
        z = math.atan2(R[1, 0], R[0, 0])
    else:
        x = math.atan2(-R[1, 2], R[1, 1])
        y = math.atan2(-R[2, 0], sy)
        z = 0

    return np.array([x, y, z])


def model(v1, coeffs):
    R = eulerAnglesToRotationMatrix(coeffs)
    RotatedV = np.matmul(R, v1.transpose())
    resultado = RotatedV.transpose()
    return resultado


def hughtransform_1p(point, Itheta, Iphi):
    p = (
        point[0] * np.cos(Itheta) * np.sin(Iphi)
        + point[1] * np.sin(Itheta) * np.sin(Iphi)
        + point[2] * np.cos(Iphi)
    )
    return p


def hugh3p_inter(coefs, p1, p2, p3, P):
    v1 = hughtransform_1p((p1 - p2 - p3), coefs[0], coefs[1]) - P
    return v1


def LIDAR_Camera_Join(
    image,
    camera_matrix,
    dist_coefs,
    LIDARpoints,
    RotacionFinal,
    traslation,
    pointsize=None,
    maxdepth=None,
):
    IRotFin = RotacionFinal.transpose()
    NTPOINTS = LIDARpoints.transpose() - traslation
    CFLidarPoints = np.matmul(IRotFin, NTPOINTS).T
    xypoints = np.zeros((CFLidarPoints.shape))
    for o, ENpoints in enumerate(CFLidarPoints):
        xypp, _ = cv2.projectPoints(
            np.float32([[0.0, 0.0, 0.0]]),
            ENpoints * 0,
            ENpoints,
            camera_matrix,
            dist_coefs,
        )
        xypoints[o] = np.hstack([xypp.squeeze(), 1])

    imagesize = image.shape
    Validbool = (
        (xypoints[:, 0] < imagesize[1])
        & (xypoints[:, 0] > 0)
        & (xypoints[:, 1] < imagesize[0])
        & (xypoints[:, 1] > 0)
    )
    validpoints = np.int16(xypoints[Validbool])
    validdepths = CFLidarPoints[Validbool]

    LIDARIMAGE = np.zeros((imagesize), dtype=np.uint8)
    if maxdepth is None:
        ecuDEPTH = (validdepths[:, 2] - np.min(validdepths[:, 2])) / (
            np.max(validdepths[:, 2]) - np.min(validdepths[:, 2])
        )
    else:
        ecuDEPTH = np.minimum(validdepths[:, 2], maxdepth) / maxdepth
    for i, point in enumerate(validpoints):
        if pointsize is None:
            LIDARIMAGE[point[1], point[0], :] = np.uint8(
                np.multiply(cm.jet(ecuDEPTH[i]), 255)
            )[:3]
        else:
            col = np.multiply(cm.jet(ecuDEPTH[i]), 255)[:3]
            center = point[0], point[1]
            cv2.circle(LIDARIMAGE, center, pointsize, color=col, thickness=-1)

    mask = cv2.bitwise_not(LIDARIMAGE[:, :, 0])
    maskedim = cv2.bitwise_and(image, image, mask=mask)
    finalim = cv2.add(maskedim, LIDARIMAGE)
    return finalim, LIDARIMAGE


def PointProject(image, camera_matrix, Points, pointsize=None):
    xypoints = np.zeros((Points.shape))
    for o, ENpoints in enumerate(Points):
        nodepthpoints = ENpoints / ENpoints[2]
        xypoints[o] = np.int16(np.matmul(camera_matrix, nodepthpoints))

    imagesize = image.shape
    Validbool = (
        (xypoints[:, 0] < imagesize[1])
        & (xypoints[:, 0] > 0)
        & (xypoints[:, 1] < imagesize[0])
        & (xypoints[:, 1] > 0)
    )
    validpoints = np.int16(xypoints[Validbool])
    validdepths = Points[Validbool]

    LIDARIMAGE = np.zeros((imagesize), dtype=np.uint8)
    ecuDEPTH = (validdepths[:, 2] - np.min(validdepths[:, 2])) / (
        np.max(validdepths[:, 2]) - np.min(validdepths[:, 2])
    )
    for i, point in enumerate(validpoints):
        if pointsize is None:
            LIDARIMAGE[point[1], point[0], :] = np.uint8(
                np.multiply(cm.jet(ecuDEPTH[i]), 255)
            )[:3]
        else:
            col = np.multiply(cm.jet(ecuDEPTH[i]), 255)[:3]
            center = point[0], point[1]
            cv2.circle(LIDARIMAGE, center, pointsize, col, -1)

    mask = cv2.bitwise_not(LIDARIMAGE[:, :, 0])
    maskedim = cv2.bitwise_and(image, image, mask=mask)
    finalim = cv2.add(maskedim, LIDARIMAGE)
    return finalim


def getPrincipalVectors(
    A,
):  # get pricipal vectors and values of a matrix centered around (0,0,0)
    VT = np.linalg.eig(np.matmul(A.T, A))
    sort = sorted(zip(VT[0], VT[1].T.tolist()), reverse=True)
    Values, Vectors = zip(*sort)
    return Vectors, Values


def ExtrinsicTransform(
    points, rotation, traslation
):  # Transform points with given rotation/traslation
    return (np.matmul(rotation, points.transpose()) + traslation).transpose()


def Dpoint2Plane(
    point, planecoefs
):  # get minimum distance from a point to a plane
    return (
        planecoefs[0] * point[0]
        + planecoefs[1] * point[1]
        + planecoefs[2] * point[2]
        + planecoefs[3]
    ) / np.linalg.norm(planecoefs[0:3])


def DistPoint2Line(
    point, linepoint1, linepoint2=np.array([0, 0, 0])
):  # get minimum destance from a point to a line
    return np.linalg.norm(
        np.cross((point - linepoint2), (point - linepoint1))
    ) / np.linalg.norm(linepoint1 - linepoint2)


def point2azimuth(
    point,
):  # get the azimuth of a point from its x,y coordinates
    if point[1] == 0:
        return np.pi / 2 - np.sign(point[0]) * np.pi / 2
    elif point[1] > 0:
        return np.pi / 2 - np.arctan(point[0] / point[1])
    else:
        return np.pi * 3 / 2 - np.arctan(point[0] / point[1])


def point2azimuthVELO(
    point,
):  # get the azimuth of a point from its x,y coordinates
    if point[0] == 0:
        return np.pi / 2 - np.sign(point[1]) * np.pi / 2
    elif point[0] > 0:
        return np.pi / 2 - np.arctan(point[1] / point[0])
    else:
        return np.pi * 3 / 2 - np.arctan(point[1] / point[0])


def FindLaserPlanes(
    inputpoints,
    samplesize=0.02,
    goodthresh=0.9,
    overlap=3,
    disthesh=0.03,
):
    if samplesize < 1:
        s = int(len(inputpoints) * samplesize)
    else:
        s = samplesize
    newstart = int(s / overlap)
    n = 0
    savedlines = []
    savedcentroids = []
    savedpoints = []
    savedvectors = []
    savedlength = []
    foundline = False
    i = s
    first = 0
    while i < len(inputpoints):
        choisepoints = inputpoints[first:i]
        centroid = np.sum(choisepoints, 0) / len(choisepoints)
        Centeredchoisepoints = choisepoints - centroid
        if foundline == False:
            U, S, VT = np.linalg.svd(Centeredchoisepoints)
            linevector = VT[0]
        goodpointsline = [
            i + centroid
            for i in Centeredchoisepoints
            if DistPoint2Line(i, linevector) < disthesh
        ]
        if len(goodpointsline) / len(choisepoints) > goodthresh:  # good found
            if foundline == False:
                oldlinevector = linevector
            foundline = True
            if angle_b_vectors(oldlinevector, linevector) > 0.1:
                Usaved = np.unique(savedpoints, axis=0)
                savedlines.append(Usaved)
                savecentroid = np.sum(Usaved, 0) / len(Usaved)
                savedcentroids.append(savecentroid)
                savedvectors.append(oldlinevector)
                savedlength.append(
                    np.linalg.norm(savedpoints[-1] - savedpoints[0])
                )
                savedpoints = []
                foundline = False
            else:
                savedpoints = savedpoints + goodpointsline
        elif foundline == True:
            Usaved = np.unique(savedpoints, axis=0)
            savedlines.append(Usaved)
            savecentroid = np.sum(Usaved, 0) / len(Usaved)
            savedcentroids.append(savecentroid)
            savedvectors.append(oldlinevector)
            savedlength.append(
                np.linalg.norm(savedpoints[-1] - savedpoints[0])
            )
            savedpoints = []
            foundline = False
        else:
            foundline = False
        n += 1
        first = newstart * n
        i = s + first
    if foundline == True:
        Usaved = np.unique(savedpoints, axis=0)
        savedlines.append(Usaved)
        savecentroid = np.sum(Usaved, 0) / len(Usaved)
        savedcentroids.append(savecentroid)
        savedvectors.append(oldlinevector)
        savedlength.append(np.linalg.norm(savedpoints[-1] - savedpoints[0]))
    return (
        np.array(savedlines),
        np.array(savedcentroids),
        np.array(savedvectors),
        np.array(savedlength),
    )


def linemodel(x, coefs):
    return x * coefs[0] + coefs[1]


def planemodel(x, y, coefs):
    return x * coefs[0] + y * coefs[1] + coefs[2]


def lineequfit(coefs, ins, outs):
    return outs - linemodel(ins, coefs)


def planeequfit(coefs, x, y, outs):
    return outs - planemodel(x, y, coefs)


def planeverticalequfit(coefs, x, y, outs):
    return (
        abs(outs - planemodel(x, y, coefs))
        + abs(10 / (coefs[0] + 0.00001))
        + abs(10 / (coefs[1] + 0.00001))
    )


def SVDLstSqr(X, Y):
    U, s, VT = np.linalg.svd(X)
    S = np.diag(s)
    Si = np.zeros([U.shape[1], VT.shape[0]])
    Si[0 : S.shape[0], 0 : S.shape[0]] = np.diag(1 / s)
    return np.matmul(VT.T, np.matmul(Si.T, np.matmul(U.T, Y)))


def SVDRigidBodyTransform(Points1, points2):
    centroid1 = np.sum(Points1, 0) / len(Points1)
    centroid2 = np.sum(points2, 0) / len(points2)
    CenteredVector1 = (Points1 - centroid1).transpose()
    CenteredVector2 = (points2 - centroid2).transpose()

    YiT = CenteredVector2.transpose()
    Xi = CenteredVector1
    Wi = np.eye(Xi.shape[1])

    S = np.matmul(Xi, np.matmul(Wi, YiT))
    U, SigNDiag, VT = np.linalg.svd(S, full_matrices=True)
    Raro = np.eye(3)
    Raro[-1, -1] = np.linalg.det(np.matmul(VT.transpose(), U.transpose()))

    RotacionFinal = np.matmul(VT.transpose(), np.matmul(Raro, U.transpose()))
    traslation = (centroid2 - np.matmul(RotacionFinal, centroid1)).reshape(
        3, 1
    )
    return traslation, RotacionFinal


def rotation_matrix_from_vectors(vec1, vec2):
    if all(np.abs(vec1) == np.abs(vec2)):
        return np.eye(3)
    a, b = (vec1 / np.linalg.norm(vec1)).reshape(3), (
        vec2 / np.linalg.norm(vec2)
    ).reshape(3)
    v = np.cross(a, b)
    c = np.dot(a, b)
    s = np.linalg.norm(v)
    kmat = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
    rotation_matrix = np.eye(3) + kmat + kmat.dot(kmat) * ((1 - c) / (s ** 2))
    return rotation_matrix


def makecylinder(model=[0, 0, 0, 1, 0, 0, 1], length=1, dense=10):
    radius = model[6]
    X, Y, Z = model[:3]
    direction = model[3:6] / np.linalg.norm(model[3:6])
    n = np.arange(0, 360, int(360 / dense))
    height = np.arange(0, length, length / dense)
    n = np.deg2rad(n)
    x, z = np.meshgrid(n, height)
    x = x.flatten()
    z = z.flatten()
    cyl = np.vstack([np.cos(x) * radius, np.sin(x) * radius, z]).T
    rotation = rotation_matrix_from_vectors([0, 0, 1], model[3:6])
    rotatedcyl = np.matmul(rotation, cyl.T).T + np.array([X, Y, Z])
    return rotatedcyl


def makeplane(coefs, center=[0, 0, 0]):
    ll, ul = -1, 1
    step = (ul - ll) / 10
    planepoints = np.meshgrid(np.arange(ll, ul, step), np.arange(ll, ul, step))
    plane = np.array(
        [
            planepoints[0].flatten(),
            planepoints[1].flatten(),
            np.zeros(len(planepoints[0].flatten())),
        ]
    ).T
    R = rotation_matrix_from_vectors([0, 0, 1], coefs[0:3])
    return np.add(np.matmul(R, plane.T).T, center)



import numpy as np
import pclpy
import open3d as o3d
from porteratzolibs.ThreeDProcessing import pclpy_utils 
from porteratzolibs.tictoc import bench_dict


def downsample(point_cloud, leaf_size=0.005, return_idx=False):
    if return_idx:
        pcd = o3d.geometry.PointCloud()
        pcd.points = o3d.utility.Vector3dVector(point_cloud)
        _, rest, _ = pcd.voxel_down_sample_and_trace(
            voxel_size=leaf_size,
            min_bound=np.array([-10, -10, -10]),
            max_bound=np.array([10, 10, 10]),
        )
        return rest[rest != -1]
    else:
        return pclpy_utils.voxelize(point_cloud, leaf_size)


def outliers(points, min_n=6, radius=0.4, organized=True):
    _points = pclpy_utils.radius_outlier_removal(points, min_n, radius, organized)
    return _points[~np.all(np.isnan(_points), axis=1)]


def farthest_point_sample(xyz, npoint):
    N, _ = xyz.shape
    centroids = []
    distance = np.ones(N) * 1e10
    farthest = np.random.randint(0, N)
    for i in range(npoint):
        centroids.append(farthest)
        centroid = xyz[farthest, :]
        dist = np.sum((xyz - centroid) ** 2, -1)
        mask = dist < distance
        distance[mask] = dist[mask]
        farthest = int(np.where(distance == np.max(distance))[0][0])
    return centroids


def normal_filter(
        subject_cloud,
        search_radius=0.05,
        verticality_threshold=0.3,
        curvature_threshold=0.3,
        return_indexes=False,
):
    bench_dict["normals"].gstep()
    non_ground_normals = pclpy_utils.extract_normals(subject_cloud, search_radius)
    bench_dict["normals"].step("extract")
    # remove Nan points
    non_nan_mask = np.bitwise_not(np.isnan(non_ground_normals.normals[:, 0]))
    non_nan_cloud = subject_cloud[non_nan_mask]
    non_nan_normals = non_ground_normals.normals[non_nan_mask]
    non_nan_curvature = non_ground_normals.curvature[non_nan_mask]
    bench_dict["normals"].step("nan")

    # get mask by filtering verticality and curvature
    verticality = np.dot(non_nan_normals, [[0], [0], [1]])
    verticality_mask = (verticality < verticality_threshold) & (
            -verticality_threshold < verticality
    )
    curvature_mask = non_nan_curvature < curvature_threshold
    verticality_curvature_mask = verticality_mask.ravel() & curvature_mask.ravel()

    only_horizontal_points = non_nan_cloud[verticality_curvature_mask]
    bench_dict["normals"].step("filter")
    bench_dict["normals"].gstop()

    if return_indexes:
        out_index = non_nan_mask
        out_index[non_nan_mask] = verticality_curvature_mask
        return out_index
    else:
        return only_horizontal_points


def seg_normals(
        filtered_points,
        search_radius=0.05,
        normalweight=0.01,
        miter=1000,
        distance=0.01,
        rlim=[0, 0.2],
):
    indices, model = pclpy_utils.segment_normals(
        filtered_points,
        search_radius=search_radius,
        model=pclpy.pcl.sample_consensus.SACMODEL_CYLINDER,
        method=pclpy.pcl.sample_consensus.SAC_RANSAC,
        normalweight=normalweight,
        miter=miter,
        distance=distance,
        rlim=rlim,
    )
    return indices, model
    


"""
MIT License

Copyright (c) 2021 porteratzo

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import pclpy
import numpy as np


def floor_remove(
    points,
    set_max_window_size=20,
    set_slope=1.0,
    set_initial_distance=0.5,
    set_max_distance=3.0,
):
    """
    Takes a point cloud and returns 2 pointclouds, the first for non ground points and the second for ground points

    Args:
        points : np.ndarray
            (n,3) point cloud

        set_max_window_size: int
            Set the maximum window size to be used in filtering ground returns.

        set_slope: float
            Set the slope value to be used in computing the height threshold.

        set_initial_distance: float
            Set the initial height above the parameterized ground surface to be considered a ground return.

        set_max_distance: float
            Set the maximum height above the parameterized ground surface to be considered a ground return.

    Returns:
        non_ground_points.xyz : np.narray (n,3)
            3d point cloud of only non ground points

        ground.xyz : np.narray (n,3)
            3d point cloud of only ground points

    """
    PointCloud = points
    ind = pclpy.pcl.vectors.Int()
    pmf = pclpy.pcl.segmentation.ApproximateProgressiveMorphologicalFilter.PointXYZ()
    pmf.setInputCloud(PointCloud)
    pmf.setMaxWindowSize(set_max_window_size)
    pmf.setSlope(set_slope)
    pmf.setInitialDistance(set_initial_distance)
    pmf.setMaxDistance(set_max_distance)
    pmf.extract(ind)
    ext = pclpy.pcl.filters.ExtractIndices.PointXYZ()
    ground = pclpy.pcl.PointCloud.PointXYZ()
    non_ground_points = pclpy.pcl.PointCloud.PointXYZ()
    ext.setInputCloud(PointCloud)
    ext.setIndices(ind)
    ext.filter(ground)
    ext.setNegative(True)
    ext.filter(non_ground_points)

    return non_ground_points.xyz, ground.xyz


def radius_outlier_removal(points, min_n=6, radius=0.4, organized=True):
    """
    Takes a point cloud and removes points that have less than minn neigbors in a certain radius

    Args:
        points : np.ndarray
            (n,3) point cloud

        min_n: int
            Neighbor threshold to keep a point

        radius: float
            Radius of the sphere a point can be in to be considered a neighbor of our sample point

        organized: bool
            If true outlier points are set to nan instead of removing the points from the cloud


    Returns:
        filtered_point_cloud.xyz : np.narray (n,3)
            (n,3) Pointcloud with outliers removed

    """
    ror_filter = pclpy.pcl.filters.RadiusOutlierRemoval.PointXYZ()
    cloud = pclpy.pcl.PointCloud.PointXYZ(points)
    ror_filter.setInputCloud(cloud)
    ror_filter.setMinNeighborsInRadius(min_n)
    ror_filter.setRadiusSearch(radius)
    ror_filter.setKeepOrganized(organized)
    filtered_point_cloud = pclpy.pcl.PointCloud.PointXYZ()
    ror_filter.filter(filtered_point_cloud)
    return filtered_point_cloud.xyz


def extract_normals(points, search_radius=0.1):
    """
    Takes a point cloud and approximates their normals using PCA

    Args:
        points : np.ndarray
            (n,3) point cloud

        search_radius: float
            Radius of the sphere a point can be in to be considered in the calculation of a sample points' normal

    Returns:
        normals : np.narray (n,3)
            (n,3) Normal vectors corresponding to the points in the input point cloud

    """
    cloud = pclpy.pcl.PointCloud.PointXYZ(points)
    cloud_normal_estimator = pclpy.pcl.features.NormalEstimationOMP.PointXYZ_Normal()
    tree = pclpy.pcl.search.KdTree.PointXYZ()
    cloud_normal_estimator.setInputCloud(cloud)
    cloud_normal_estimator.setSearchMethod(tree)
    cloud_normal_estimator.setRadiusSearch(search_radius)
    normals = pclpy.pcl.PointCloud.Normal()
    cloud_normal_estimator.compute(normals)
    return normals


def euclidean_cluster_extract(
    points, tolerance=2, min_cluster_size=20, max_cluster_size=25000
):
    """
    Takes a point cloud and clusters the points with euclidean clustering

    Args:
        points : np.ndarray
            (n,3) point cloud

        tolerance: int
            Maximum distance a point can be to a cluster to added to that cluster

        min_cluster_size: int
            Minimum number of points a cluster must have to be returned

        max_cluster_size: int
            Maximum number of points a cluster must have to be returned


    Returns:
        cluster_list : list
            List of (n,3) Pointclouds representing each cluster

    """
    filtered_points = pclpy.pcl.segmentation.EuclideanClusterExtraction.PointXYZ()
    kd_tree = pclpy.pcl.search.KdTree.PointXYZ()
    points_to_cluster = pclpy.pcl.PointCloud.PointXYZ(points)

    kd_tree.setInputCloud(points_to_cluster)
    filtered_points.setInputCloud(points_to_cluster)
    filtered_points.setClusterTolerance(tolerance)
    filtered_points.setMinClusterSize(min_cluster_size)
    filtered_points.setMaxClusterSize(max_cluster_size)
    filtered_points.setSearchMethod(kd_tree)

    point_indexes = pclpy.pcl.vectors.PointIndices()
    filtered_points.extract(point_indexes)

    cluster_list = [points_to_cluster.xyz[i2.indices] for i2 in point_indexes]
    return cluster_list


def region_growing(
    Points, ksearch=30, minc=20, maxc=100000, nn=30, smoothness=30.0, curvature=1.0
):
    """
    Takes a point cloud and clusters the points with region growing

    Args:
        points : np.ndarray
            (n,3) point cloud

        Ksearch: int
            Number of points used to estimate a points normal

        minc: int
            Minimum number of points a cluster must have to be returned

        maxc: int
            Maximum number of points a cluster must have to be returned

        nn: int
            Number of nearest neighbors used by the region growing algorithm

        smoothness:
            Smoothness threshold used in region growing

        curvature:
            Curvature threshold used in region growing

    Returns:
        region_growing_clusters: list
            list of (n,3) Pointclouds representing each cluster

    """
    pointcloud = pclpy.pcl.PointCloud.PointXYZ(Points)
    pointcloud_normals = pclpy.pcl.features.NormalEstimation.PointXYZ_Normal()
    tree = pclpy.pcl.search.KdTree.PointXYZ()

    pointcloud_normals.setInputCloud(pointcloud)
    pointcloud_normals.setSearchMethod(tree)
    pointcloud_normals.setKSearch(ksearch)
    normals = pclpy.pcl.PointCloud.Normal()
    pointcloud_normals.compute(normals)

    region_growing_clusterer = pclpy.pcl.segmentation.RegionGrowing.PointXYZ_Normal()
    region_growing_clusterer.setInputCloud(pointcloud)
    region_growing_clusterer.setInputNormals(normals)
    region_growing_clusterer.setMinClusterSize(minc)
    region_growing_clusterer.setMaxClusterSize(maxc)
    region_growing_clusterer.setSearchMethod(tree)
    region_growing_clusterer.setNumberOfNeighbours(nn)
    region_growing_clusterer.setSmoothnessThreshold(smoothness / 180.0 * np.pi)
    region_growing_clusterer.setCurvatureThreshold(curvature)

    clusters = pclpy.pcl.vectors.PointIndices()
    region_growing_clusterer.extract(clusters)

    region_growing_clusters = [pointcloud.xyz[i2.indices] for i2 in clusters]
    return region_growing_clusters


def segment(
    points,
    model=pclpy.pcl.sample_consensus.SACMODEL_LINE,
    method=pclpy.pcl.sample_consensus.SAC_RANSAC,
    miter=1000,
    distance=0.5,
    rlim=[0, 0.5],
):
    """
    Takes a point cloud and removes points that have less than minn neigbors in a certain radius

    Args:
        points : np.ndarray
            (n,3) point cloud

        model: int
            A pclpy.pcl.sample_consensus.MODEL value representing a ransac model

        method: float
            pclpy.pcl.sample_consensus.METHOD to use

        miter: bool
            Maximum iterations for ransac

        distance:
            Maximum distance a point can be from the model

        rlim:
            Radius limit for cylinder model


    Returns:
        pI.indices: np.narray (n)
            Indices of points that fit the model

        Mc.values: np.narray (n)
            Model coefficients

    """
    pointcloud = pclpy.pcl.PointCloud.PointXYZ(points)
    segmenter = pclpy.pcl.segmentation.SACSegmentation.PointXYZ()

    segmenter.setInputCloud(pointcloud)
    segmenter.setDistanceThreshold(distance)
    segmenter.setOptimizeCoefficients(True)
    segmenter.setMethodType(method)
    segmenter.setModelType(model)
    segmenter.setMaxIterations(miter)
    segmenter.setRadiusLimits(rlim[0], rlim[1])
    pI = pclpy.pcl.PointIndices()
    Mc = pclpy.pcl.ModelCoefficients()
    segmenter.segment(pI, Mc)
    return pI.indices, Mc.values


def segment_normals(
    points,
    search_radius=20,
    model=pclpy.pcl.sample_consensus.SACMODEL_LINE,
    method=pclpy.pcl.sample_consensus.SAC_RANSAC,
    normalweight=0.0001,
    miter=1000,
    distance=0.5,
    rlim=[0, 0.5],
):
    """
    Takes a point cloud and removes points that have less than minn neigbors in a certain radius

    Args:
        points : np.ndarray
            (n,3) point cloud

        search_radius: float
            Radius of the sphere a point can be in to be considered in the calculation of a sample points' normal

        model: int
            A pclpy.pcl.sample_consensus.MODEL value representing a ransac model

        method: float
            pclpy.pcl.sample_consensus.METHOD to use

        normalweight:
            Normal weight for ransacfromnormals

        miter: bool
            Maximum iterations for ransac

        distance:
            Maximum distance a point can be from the model

        rlim:
            Radius limit for cylinder model


    Returns:
        pI.indices: np.narray (n)
            Indices of points that fit the model

        Mc.values: np.narray (n)
            Model coefficients

    """
    pointcloud_normals = extract_normals(points, search_radius)

    pointcloud = pclpy.pcl.PointCloud.PointXYZ(points)
    segmenter = pclpy.pcl.segmentation.SACSegmentationFromNormals.PointXYZ_Normal()

    segmenter.setInputCloud(pointcloud)
    segmenter.setInputNormals(pointcloud_normals)
    segmenter.setDistanceThreshold(distance)
    segmenter.setOptimizeCoefficients(True)
    segmenter.setMethodType(method)
    segmenter.setModelType(model)
    segmenter.setMaxIterations(miter)
    segmenter.setRadiusLimits(rlim[0], rlim[1])
    segmenter.setNormalDistanceWeight(normalweight)
    pI = pclpy.pcl.PointIndices()
    Mc = pclpy.pcl.ModelCoefficients()
    segmenter.segment(pI, Mc)
    return pI.indices, Mc.values


def findstemsLiDAR(pointsXYZ):
    """
    Takes a point cloud from a Cylindrical LiDAR and extract stems and their models

    Args:
        points : np.ndarray
            (n,3) point cloud

    Returns:
        stemsR : list(np.narray (n,3))
            List of (n,3) Pointclouds belonging to each stem

        models : list(np.narray (n))
            List of model coefficients corresponding to each extracted stem

    """
    non_ground_points, ground = floor_remove(pointsXYZ)
    flatpoints = np.hstack(
        [non_ground_points[:, 0:2], np.zeros_like(non_ground_points)[:, 0:1]]
    )

    filtered_points = radius_outlier_removal(flatpoints)
    notgoodpoints = non_ground_points[np.isnan(filtered_points[:, 0])]
    goodpoints = non_ground_points[np.bitwise_not(np.isnan(filtered_points[:, 0]))]

    cluster_list = euclidean_cluster_extract(goodpoints)
    rg_clusters = []
    for i in cluster_list:
        rg_clusters.append(region_growing(i))

    models = []
    stem_clouds = []
    for i in rg_clusters:
        for p in i:
            indices, model = segment_normals(p)
            prop = len(p[indices]) / len(p)
            if (
                len(indices) > 1
                and prop > 0.0
                and np.arccos(np.dot([0, 0, 1], model[3:6])) < 0.6
            ):
                points = p[indices]
                PC, _, _ = Plane.getPrincipalComponents(points)
                if PC[0] / PC[1] > 10:
                    stem_clouds.append(points)
                    models.append(model)
    return stem_clouds, models


def voxelize(points, leaf=0.1):
    """
    Use voxelgrid to subsample a pointcloud

    Args:
        points : np.ndarray
            (n,3) point cloud

        leaf: float
            Voxelsize

    Returns:
        VFmm: np.narray (n,3)
            (n,3) subsampled Pointcloud

    """
    if type(points) == pclpy.pcl.PointCloud.PointXYZRGB:
        cloud = points
        voxel_filter = pclpy.pcl.filters.VoxelGrid.PointXYZRGB()
        filtered_pointcloud = pclpy.pcl.PointCloud.PointXYZRGB()
    else:
        cloud = pclpy.pcl.PointCloud.PointXYZ(points)
        voxel_filter = pclpy.pcl.filters.VoxelGrid.PointXYZ()
        filtered_pointcloud = pclpy.pcl.PointCloud.PointXYZ()

    voxel_filter.setLeafSize(leaf, leaf, leaf)
    voxel_filter.setInputCloud(cloud)

    voxel_filter.filter(filtered_pointcloud)
    if type(points) == pclpy.pcl.PointCloud.PointXYZRGB:
        return filtered_pointcloud
    else:
        return filtered_pointcloud.xyz



# General
import cv2
import sys
import numpy as np
import imp
import tensorflow as tf
import os

import threading
import datetime
import time
import glob

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)

# Script para codigo terminado

# Probar rendimiento
timestart = [time.perf_counter() for i in range(5)]


def tic(number=0):
    global timestart
    timestart[number] = time.perf_counter()


def toc(number=0):
    global timestart
    return time.perf_counter() - timestart[number]


class tictocclass:
    def __init__(self):
        self.tictocTimer = time.perf_counter()

    def tic(self):
        self.tictocTimer = time.perf_counter()

    def toc(self):
        return time.perf_counter() - self.tictocTimer


def flushcam(cam, Sdelay=0.001):
    framesWithDelayCount = 0
    while framesWithDelayCount <= 1:
        tic()
        cam.grab()
        delay = toc()
        if delay > Sdelay:
            framesWithDelayCount += 1


def load_model_EfficientDet(path, automl=False):
    model = tf.saved_model.load(path)
    if automl:
        model = model.signatures["serving_default"]
    return model


class Lite_Detector(tf.lite.Interpreter):
    def __init__(
        self,
        model_path="coco_ssd_mobilenet_v1_1.0_quant_2018_06_29/detect.tflite",
        efficient=False,
    ):
        super().__init__(model_path)
        self.isEfficient = efficient
        self.allocate_tensors()
        self.input_details = self.get_input_details()
        self.output_details = self.get_output_details()
        self.labels = load_labels(True)
        self.threshhold = 0.6
        self.sub_128 = False
        self.div_128 = False

    def set_threshhold(self, threshhold):
        self.threshhold = threshhold

    def set_normalize(self, sub_128, div_128):
        self.sub_128 = sub_128
        self.div_128 = div_128

    def detect(self, image):
        if self.sub_128:
            image = image.copy() - 128
        if self.div_128:
            image = image.copy() / 128
        input_data = self.input_details[0]["dtype"](
            np.expand_dims(
                cv2.resize(image, tuple(self.input_details[0]["shape"][1:3])),
                axis=0,
            )
        )
        self.set_tensor(self.input_details[0]["index"], input_data)
        self.invoke()
        output_dict = {}
        if self.isEfficient:
            partdict = self.get_tensor(self.output_details[0]["index"])[0]
            output_dict["detection_classes"] = np.int8(partdict[:, 6])
            output_dict["detection_boxes"] = np.int16(partdict[:, 1:5])
            output_dict["detection_scores"] = partdict[:, 5]
        else:
            output_dict["detection_classes"] = (
                np.int8(self.get_tensor(self.output_details[1]["index"])[0])
                + 1
            )
            output_dict["detection_boxes"] = self.get_tensor(
                self.output_details[0]["index"]
            )[0]
            output_dict["detection_scores"] = self.get_tensor(
                self.output_details[2]["index"]
            )[0]

        self.output_dict = output_dict
        return output_dict

    def detectEfficientNet(self, image):
        input_data = self.input_details[0]["dtype"](
            np.expand_dims(
                cv2.resize(image, tuple(self.input_details[0]["shape"][1:3])),
                axis=0,
            )
        )
        self.set_tensor(self.input_details[0]["index"], input_data)
        self.invoke()
        partdict = modelLite.get_tensor(output_details[0]["index"])[0]
        output_dict = {}
        output_dict["detection_classes"] = np.int8(partdict[:, 6])
        output_dict["detection_boxes"] = np.int16(partdict[:, 1:5])
        output_dict["detection_scores"] = partdict[:, 5]

    def Draw(self, image):
        if self.isEfficient:
            outputFrame = detect_all_efficient_Lite(
                image,
                self.output_dict,
                self.labels,
                self.input_details[0]["shape"][1],
                self.threshhold,
            )
        else:
            outputFrame = detect_all(
                image, self.output_dict, self.labels, self.threshhold
            )
        input_details[0]["shape"][1]
        return outputFrame

    def DetectandDraw(self, image):
        self.detect(image)
        if self.isEfficient:
            outputFrame = detect_all_efficient_Lite(
                image,
                self.output_dict,
                self.labels,
                self.input_details[0]["shape"][1],
                self.threshhold,
            )
        else:
            outputFrame = detect_all(
                image, self.output_dict, self.labels, self.threshhold
            )
        return outputFrame

    def FilterbyClass(self, classes):
        assert type(classes) is list, "classes shoud be a list"
        classfind = [
            self.output_dict["detection_classes"] == i for i in classes
        ]
        classinter = [
            any([p[i] for p in classfind]) for i in range(len(classfind[0]))
        ]
        intersection = np.bitwise_and(
            self.output_dict["detection_scores"] > self.threshhold, classinter
        )
        self.output_dict["detection_classes"] = self.output_dict[
            "detection_classes"
        ][intersection]
        self.output_dict["detection_boxes"] = self.output_dict[
            "detection_boxes"
        ][intersection]
        self.output_dict["detection_scores"] = self.output_dict[
            "detection_scores"
        ][intersection]


# Tomar salida de red y dibujar personas
def detect_people(image, output_dict, thresh=0.50):
    imageBox = image.copy()
    h, w, _ = image.shape
    gooddetections = np.argwhere(output_dict["detection_scores"] > thresh)
    rightcat = np.argwhere(output_dict["detection_classes"] == 1)
    intersects = np.intersect1d(gooddetections, rightcat)
    for i in intersects:
        y1, x1, y2, x2 = output_dict["detection_boxes"][i]
        x1, y1, x2, y2 = [
            int(imageBox.shape[1] * x1),
            int(imageBox.shape[0] * y1),
            int(imageBox.shape[1] * x2),
            int(imageBox.shape[0] * y2),
        ]
        x1, y1, x2, y2 = max(x1, 0), max(y1, 0), min(x2, w), min(y2, h)
        cv2.rectangle(imageBox, (x1, y1), (x2, y2), (0, 0, 255), 2)

        cv2.putText(
            imageBox,
            str(int(output_dict["detection_scores"][i] * 100)) + "%",
            (x1, y2 - 30),
            cv2.FONT_HERSHEY_SIMPLEX,
            0.8,
            (255, 255, 255),
            2,
        )
        cv2.putText(
            imageBox,
            str(int(output_dict["detection_scores"][i] * 100)) + "%",
            (x1, y2 - 10),
            cv2.FONT_HERSHEY_SIMPLEX,
            0.8,
            (0, 0, 0),
            2,
        )
    return imageBox


def detect_cars(image, output_dict, thresh=0.50):
    imageBox = image.copy()
    h, w, _ = image.shape
    gooddetections = np.argwhere(output_dict["detection_scores"] > thresh)
    rightcat = np.argwhere(output_dict["detection_classes"] == 3)
    intersects = np.intersect1d(gooddetections, rightcat)
    for i in intersects:
        y1, x1, y2, x2 = output_dict["detection_boxes"][i]
        x1, y1, x2, y2 = [
            int(imageBox.shape[1] * x1),
            int(imageBox.shape[0] * y1),
            int(imageBox.shape[1] * x2),
            int(imageBox.shape[0] * y2),
        ]
        x1, y1, x2, y2 = max(x1, 0), max(y1, 0), min(x2, w), min(y2, h)
        cv2.rectangle(imageBox, (x1, y1), (x2, y2), (0, 0, 255), 2)

        cv2.putText(
            imageBox,
            str(int(output_dict["detection_scores"][i] * 100)) + "%",
            (x1, y2 - 30),
            cv2.FONT_HERSHEY_SIMPLEX,
            0.8,
            (255, 255, 255),
            2,
        )
        cv2.putText(
            imageBox,
            str(int(output_dict["detection_scores"][i] * 100)) + "%",
            (x1, y2 - 10),
            cv2.FONT_HERSHEY_SIMPLEX,
            0.8,
            (0, 0, 0),
            2,
        )
    return imageBox


def getcrop(
    image,
    output_dict,
    thresh=0.50,
    ythresh1=0,
    ythresh2=1,
    xthresh1=0,
    xthresh2=1,
):
    imageBox = image.copy()
    h, w, _ = image.shape
    gooddetections = np.argwhere(output_dict["detection_scores"] > thresh)
    rightcat1 = np.argwhere(output_dict["detection_classes"] == 3)
    # rightcat2 = np.argwhere(output_dict['detection_classes']==6)
    rightcat3 = np.argwhere(output_dict["detection_classes"] == 8)
    rightcat = np.concatenate([rightcat1, rightcat3])
    intersects = np.intersect1d(gooddetections, rightcat)
    images = []
    for i in intersects:
        y1, x1, y2, x2 = output_dict["detection_boxes"][i]
        x1, y1, x2, y2 = [
            int(imageBox.shape[1] * x1),
            int(imageBox.shape[0] * y1),
            int(imageBox.shape[1] * x2),
            int(imageBox.shape[0] * y2),
        ]

        if (
            (y1 > ythresh1 * h)
            & (y2 < ythresh2 * h)
            & (x1 > xthresh1 * w)
            & (x2 < xthresh2 * w)
        ):
            x1, y1, x2, y2 = max(x1, 0), max(y1, 0), min(x2, w), min(y2, h)
            images.append(imageBox[y1:y2, x1:x2])
    return images


# Tomar salida de red y dibujar objetos detectados
def detect_all(image, output_dict, labels_dict, thresh=0.50):
    imageBox = image.copy()
    h, w, _ = image.shape
    gooddetections = np.argwhere(
        output_dict["detection_scores"] > thresh
    ).ravel()
    for i in gooddetections:
        y1, x1, y2, x2 = output_dict["detection_boxes"][i]
        x1, y1, x2, y2 = [
            int(imageBox.shape[1] * x1),
            int(imageBox.shape[0] * y1),
            int(imageBox.shape[1] * x2),
            int(imageBox.shape[0] * y2),
        ]
        x1, y1, x2, y2 = max(x1, 0), max(y1, 0), min(x2, w), min(y2, h)
        cv2.rectangle(imageBox, (x1, y1), (x2, y2), (0, 0, 255), 2)

        cv2.putText(
            imageBox,
            str(labels_dict[str(output_dict["detection_classes"][i])])
            + ": "
            + str(int(output_dict["detection_scores"][i] * 100))
            + "%",
            (x1, y2 - 30),
            cv2.FONT_HERSHEY_SIMPLEX,
            0.8,
            (255, 255, 255),
            2,
        )
        cv2.putText(
            imageBox,
            str(labels_dict[str(output_dict["detection_classes"][i])])
            + ": "
            + str(int(output_dict["detection_scores"][i] * 100))
            + "%",
            (x1, y2 - 10),
            cv2.FONT_HERSHEY_SIMPLEX,
            0.8,
            (0, 0, 0),
            2,
        )
    return imageBox


def detect_all_efficient_Lite(
    image, output_dict, labels_dict, scale=300, thresh=0.50
):
    imageBox = image.copy()
    h, w, _ = image.shape
    gooddetections = np.argwhere(
        output_dict["detection_scores"] > thresh
    ).ravel()
    for i in gooddetections:
        y1, x1, y2, x2 = output_dict["detection_boxes"][i]
        x1, y1, x2, y2 = [
            int(imageBox.shape[1] / scale * x1),
            int(imageBox.shape[0] / scale * y1),
            int(imageBox.shape[1] / scale * x2),
            int(imageBox.shape[0] / scale * y2),
        ]
        x1, y1, x2, y2 = max(x1, 0), max(y1, 0), min(x2, w), min(y2, h)
        cv2.rectangle(imageBox, (x1, y1), (x2, y2), (0, 0, 255), 2)

        cv2.putText(
            imageBox,
            str(labels_dict[str(output_dict["detection_classes"][i])])
            + ": "
            + str(int(output_dict["detection_scores"][i] * 100))
            + "%",
            (x1, y2 - 30),
            cv2.FONT_HERSHEY_SIMPLEX,
            0.8,
            (255, 255, 255),
            2,
        )
        cv2.putText(
            imageBox,
            str(labels_dict[str(output_dict["detection_classes"][i])])
            + ": "
            + str(int(output_dict["detection_scores"][i] * 100))
            + "%",
            (x1, y2 - 10),
            cv2.FONT_HERSHEY_SIMPLEX,
            0.8,
            (0, 0, 0),
            2,
        )
    return imageBox


def load_labels(lite=False):
    if lite:
        PATH_TO_LABELS = r"labelmap.txt"
    else:
        PATH_TO_LABELS = "mscoco_label_map.pbtxt"

    with open(PATH_TO_LABELS, "r") as fh:
        graph_str = fh.read()

    if lite:
        Labelsf = graph_str.split("\n")
    else:
        Labelsf = graph_str.split("item")

    MyLabels = {}
    if lite:
        for n, i in enumerate(Labelsf):
            MyLabels[str(n)] = i

    else:
        for index in Labelsf:
            pos1id = index.find("id: ") + 4
            pos2id = pos1id + 2
            if not index[pos1id:pos2id].isdigit():
                pos2id -= 1
            pos1dn = index.find("display_name: ") + 15
            pos2dn = index.find('"', pos1dn)
            MyLabels[index[pos1id:pos2id]] = index[pos1dn:pos2dn]

    return MyLabels


# Tomar salida de red y dibujar objetos detectados
def detect_all_EfficientDet(image, output_dict, labels_dict, thresh=0.50):
    imageBox = image.copy()
    for i in output_dict["detections:0"][0]:
        if i[5] > thresh:
            y1, x1, y2, x2 = np.int32(i[1:5].numpy())
            color = (0, 255, 0)
            if i[6].numpy() == 1:
                color = (0, 0, 255)
            cv2.rectangle(imageBox, (x1, y1), (x2, y2), color, 2)
            cv2.putText(
                imageBox,
                str(labels_dict[str(int(i[6].numpy()))])
                + ": "
                + str(int(i[5].numpy() * 100))
                + "%",
                (x2 - 50, y2),
                cv2.FONT_HERSHEY_SIMPLEX,
                0.8,
                (0, 0, 0),
                2,
            )
    return imageBox


# Correr red sobre una imagen
def run_inference_for_single_image_EfficientDet(model, image, automl=False):
    image = np.asarray(image)

    if automl:
        input_tensor = tf.convert_to_tensor(image)
        input_tensor = input_tensor[tf.newaxis, ...]
        outputdictPRE = model(input_tensor)
        outputdict = {}
        outputdict["detection_boxes"] = outputdictPRE["detections:0"][0][
            :, 1:5
        ].numpy()
        outputdict["detection_boxes"][:, 0] = (
            outputdict["detection_boxes"][:, 0] / image.shape[0]
        )
        outputdict["detection_boxes"][:, 1] = (
            outputdict["detection_boxes"][:, 1] / image.shape[1]
        )
        outputdict["detection_boxes"][:, 2] = (
            outputdict["detection_boxes"][:, 2] / image.shape[0]
        )
        outputdict["detection_boxes"][:, 3] = (
            outputdict["detection_boxes"][:, 3] / image.shape[1]
        )
        outputdict["detection_classes"] = np.int8(
            outputdictPRE["detections:0"][0][:, 6].numpy()
        )
        outputdict["detection_scores"] = outputdictPRE["detections:0"][0][
            :, 5
        ].numpy()
    else:
        input_tensor = tf.convert_to_tensor(image)
        input_tensor = input_tensor[tf.newaxis, ...]
        outputdict = model(input_tensor)
        outputdict["detection_boxes"] = outputdict["detection_boxes"][0]

        outputdict["detection_classes"] = np.int8(
            outputdict["detection_classes"][0]
        )
        outputdict["detection_scores"] = outputdict["detection_scores"][0]

    return outputdict



from matplotlib import pyplot as plt
from matplotlib import cm
import numpy as np
import glob
import pandas as pd
import open3d
import scipy
import scipy.linalg
import scipy.optimize
from mpl_toolkits.mplot3d import Axes3D
import math
import os
from sklearn import linear_model
import TreeToolML.porteratzolibs.Corners as Corners
import pdb
from pdb import set_trace as Debug
from scipy.ndimage import gaussian_filter1d
import time
from sklearn.cluster import DBSCAN

timestart = time.perf_counter()

def tic():
    global timestart
    timestart = time.perf_counter()


def toc():
    global timestart
    return time.perf_counter() - timestart


def newnormalaxis(points, normal1, normal2, normal3):
    matrix = np.vstack([normal1, normal2, normal3])
    newpoints = np.matmul(matrix, points.T)
    return newpoints.T


def findconicsection(points):
    ecuation = []
    for i in points:
        ecuation.append([i[0] ** 2, i[0] * i[1], i[1] ** 2, i[0], i[1], 1])
    ecuation = np.array(ecuation)
    VT = np.linalg.eig(np.matmul(ecuation.T, ecuation))
    eigval = VT[0]
    sort = sorted(zip(eigval, VT[1].T.tolist()), reverse=True)
    _, VT = zip(*sort)
    MF = VT[-1]
    return MF


def conic_center(a):
    a, b, c, d, e, f = a[0], a[1] / 2, a[2], a[3] / 2, a[4] / 2, a[5]
    num = a * c - b * b
    x0 = (b * e - c * d) / num
    y0 = (b * d - a * e) / num
    return np.array([x0, y0])


def ellipse_angle_of_rotation2(a):
    b, c, d, f, g, a = a[1] / 2, a[2], a[3] / 2, a[4] / 2, a[5], a[0]
    if b == 0:
        if a > c:
            return 0
        else:
            return np.pi / 2
    else:
        if a > c:
            return np.arctan(2 * b / (a - c)) / 2
        else:
            return np.pi / 2 + np.arctan(2 * b / (a - c)) / 2


def ellipse_axis_length(a):
    b, c, d, f, g, a = a[1] / 2, a[2], a[3] / 2, a[4] / 2, a[5], a[0]
    up = 2 * (a * f * f + c * d * d + g * b * b - 2 * b * d * f - a * c * g)
    down1 = (b * b - a * c) * (
        (c - a) * np.sqrt(1 + 4 * b * b / ((a - c) * (a - c))) - (c + a)
    )
    down2 = (b * b - a * c) * (
        (a - c) * np.sqrt(1 + 4 * b * b / ((a - c) * (a - c))) - (c + a)
    )
    res1 = np.sqrt(up / down1)
    res2 = np.sqrt(up / down2)
    return np.array([res1, res2])


def make2d(points, thresh=0.04):
    goodpoints, _ = Corners.FitPlanePoints(points, thresh=thresh)
    centroid = np.sum(goodpoints, 0) / len(goodpoints)
    cgoodpoints = goodpoints - centroid
    VT = np.linalg.eig(np.matmul(cgoodpoints.T, cgoodpoints))
    eigval = VT[0]
    sort = sorted(zip(eigval, VT[1].T.tolist()), reverse=True)
    _, VT = zip(*sort)
    newaxisCenteredchoisepoints = newnormalaxis(
        cgoodpoints, VT[0], VT[1], VT[2]
    )
    return newaxisCenteredchoisepoints


def make2dnoRAN(goodpoints):
    centroid = np.sum(goodpoints, 0) / len(goodpoints)
    cgoodpoints = goodpoints - centroid
    VT = np.linalg.eig(np.matmul(cgoodpoints.T, cgoodpoints))
    eigval = VT[0]
    sort = sorted(zip(eigval, VT[1].T.tolist()), reverse=True)
    _, VT = zip(*sort)
    newaxisCenteredchoisepoints = newnormalaxis(
        cgoodpoints, VT[0], VT[1], VT[2]
    )
    return newaxisCenteredchoisepoints


def testconicpoint(point, coef):
    i = point
    ecuation = [i[0] ** 2, i[0] * i[1], i[1] ** 2, i[0], i[1], 1]
    res = np.matmul(ecuation, coef)
    return res


def findclosest(point, points):
    lengths = []
    index = np.arange(len(points))
    for i in points:
        lengths.append(Corners.point_dis(point, i))
    out = [x for _, x in sorted(zip(lengths, index))]
    return out


def findPlanePointswithcentroid(coefs, centroid, points):
    out = []
    newcoefs = [coefs[0], coefs[1], coefs[2], 0]
    newpoints = np.subtract(points, centroid)
    for i in newpoints:
        out.append(Corners.Dpoint2Plane(i, newcoefs))
    return out


def makepointvector(coefs, centroid=[0, 0, 0], length=1, dense=10):
    assert len(coefs) == 3, "Need x,y,z normalvector"
    newcoefs = np.array(coefs) / np.linalg.norm(np.array(coefs))
    pointline = np.arange(
        length / dense, length + length / dense, length / dense
    )
    pointline = np.vstack([pointline, pointline, pointline])
    out = np.add(np.multiply(pointline.T, newcoefs), centroid)
    return out


def getPrincipalComponents(A):
    centroid = np.mean(A, 0)
    A = np.subtract(A, centroid)
    VT = np.linalg.eig(np.matmul(A.T, A))
    eigval = VT[0]
    sort = sorted(zip(eigval, VT[1].T.tolist()), reverse=True)
    Val, VT = zip(*sort)
    return Val, VT, centroid


def findconictype(coefs, voc=True):
    a, b, c, d, e, f = coefs
    re = b * b - 4 * a * c
    if re < 0:
        if a == c and b == 0:
            if voc:
                print("circle")
            nu = 0
        else:
            if voc:
                print("ellipse")
            nu = 1
    elif re == 0:
        if voc:
            print("parabola")
        nu = 2
    else:
        if a + c != 0:
            if voc:
                print("hyperbola")
            nu = 3
        else:
            if voc:
                print("rectangular hyperbola")
            nu = 4


def FindArcPlanes(
    inputpoints, samplesize=0.02, goodthresh=0.9, overlap=3, disthesh=0.03
):
    if samplesize < 1:
        s = int(len(inputpoints) * samplesize)
    else:
        s = samplesize

    newstart = int(s / overlap)
    n = 0
    savedlines = []
    savedcentroids = []
    savedpoints = []
    savedlength = []
    savednormals = []
    foundline = False
    i = s
    first = 0
    prethreshold = goodthresh
    while i < len(inputpoints):
        choisepoints = inputpoints[first:i]
        flatpoints = make2dnoRAN(choisepoints)
        coefs = findconicsection(flatpoints)
        buff = [
            [i2, k]
            for i2, k in zip(flatpoints, choisepoints)
            if abs(testconicpoint(i2, coefs)) < disthesh
        ]
        if len(buff) / len(choisepoints) > prethreshold:  # good found
            # goodpointsflat =[]
            # goodopoints =[]
            # for i3 in buff:
            #    goodpointsflat.append(i3[0])
            #    goodopoints.append(i3[1])
            goodpointsflat, goodopoints = zip(*buff)
            savedpoints = goodopoints
            savedlines.append(savedpoints)
            savecentroid = np.sum(savedpoints, 0) / len(savedpoints)
            savedcentroids.append(savecentroid)
            A = goodopoints - savecentroid
            VT = np.linalg.eig(np.matmul(A.T, A))
            eigval = VT[0]
            sort = sorted(zip(eigval, VT[1].T.tolist()), reverse=True)
            _, VT = zip(*sort)
            savednormals.append([VT[0], VT[2]])
            savedlength.append(
                np.linalg.norm(savedpoints[-1] - savedpoints[0])
            )
        n += 1
        first = newstart * n
        i = s + first
    return (
        np.array(savedlines),
        np.array(savedcentroids),
        np.array(savedlength),
        np.array(savednormals),
    )


def GetGradient(x):
    [X, Y, Z] = np.split(x, 3, 1)
    X = X.squeeze()
    Y = Y.squeeze()
    Z = Z.squeeze()
    derivativeX = np.gradient(X)
    derivativeY = np.gradient(Y)
    derivativeZ = np.gradient(Z)
    derivative = np.vstack([derivativeX, derivativeY, derivativeZ]).T
    return derivative


def gaussian(x, sig):
    return (
        1.0
        / (sig * np.sqrt(2.0 * np.pi))
        * np.exp(-np.power(x / sig, 2.0) / 2)
    )


def GaussinaKernel1D(size=5, sigma=1, step=1):
    side = size // 2
    val = gaussian(np.arange(-side, side + 1, step), sigma)
    return val / np.sum(val)


def Der2gaussian(x, sig):
    return (
        -(sig ** 2 - x ** 2)
        / ((sig ** 5) * np.sqrt(2.0 * np.pi))
        * np.exp(-np.power(x / sig, 2.0) / 2)
    )


def Der2GaussinaKernel1D(size=5, sigma=1, step=1):
    side = size // 2
    val = Der2gaussian(np.arange(-side, side + 1, step), sigma)
    return val


def bilateral(linepoints, bound=9, sigInt=0.03):
    side = bound // 2
    matvector = makesignalmatrix(linepoints, bound)
    gausserInt = lambda t: gaussian(t, sigInt)
    funcGausInt = np.vectorize(gausserInt)
    IntensityPart = np.abs(np.subtract(matvector, linepoints))
    Intensitykernel = funcGausInt(IntensityPart)
    W = np.sum(Intensitykernel, axis=0)
    Upper = np.sum(Intensitykernel * matvector, axis=0)
    return Upper / W


def makesignalmatrix(signal, bound, stride=1):
    assert bound % 2 == 1, "bound must be odd"
    side = bound // 2
    padded = np.concatenate([np.zeros([side]), signal, np.zeros([side])])
    mat = [np.arange(i, len(signal) + i, stride) for i in range(bound)]
    return padded[np.array(mat)]


def Filter3dline(points, sigma=2):
    X = gaussian_filter1d(points[:, 0], sigma)
    Y = gaussian_filter1d(points[:, 1], sigma)
    Z = gaussian_filter1d(points[:, 2], sigma)
    linepointsFiltered = np.vstack([X, Y, Z]).T
    return linepointsFiltered


def bilateral3dline(points, bound, sigma=2):
    X = bilateral(points[:, 0], bound, sigma)
    Y = bilateral(points[:, 1], bound, sigma)
    Z = bilateral(points[:, 2], bound, sigma)
    linepointsFiltered = np.vstack([X, Y, Z]).T
    return linepointsFiltered


def bilateral3d(linepoints, bound=9, sigInt=0.03):
    side = bound // 2
    out = []

    gausserInt = lambda t: gaussian(t, sigInt)
    funcGausInt = np.vectorize(gausserInt)
    matvector = makesignalmatrix(linepoints, bound)

    IntensityPart = np.linalg.norm(matvector - linepoints, axis=1)
    Intensitykernel = funcGausInt(IntensityPart)
    W = np.sum(Intensitykernel, axis=0)
    Upper = np.sum(Intensitykernel * matvector, axis=0)

    for n, i in enumerate(linepoints):
        position = n + side
        positionpoints = matvector[position - side : position + side + 1]
        IntensityPart = np.linalg.norm(
            positionpoints - matvector[position], axis=1
        )
        Intensitykernel = funcGausInt(IntensityPart)
        buf = Intensitykernel
        W = np.sum(buf)
        Upper = np.sum(buf * positionpoints.T, axis=1)
        out.append(Upper / W)
    return np.array(out)


def bilateral3dNew(linepoints, bound=9, sigInt=0.03):
    side = bound // 2
    matvector = makesignalmatrix3D(linepoints, bound)
    IntensityPart = np.linalg.norm(
        np.stack(
            [
                matvector[0, :, :] - linepoints[0, :],
                matvector[1, :, :] - linepoints[1, :],
                matvector[2, :, :] - linepoints[2, :],
            ]
        ),
        axis=0,
    )
    Intensitykernel = gaussian(IntensityPart, sigInt)
    W = np.sum(Intensitykernel, axis=0)
    Upper = np.sum(Intensitykernel * matvector, axis=1)
    return (Upper / W).T


def makesignalmatrix3D(signal, bound, stride=1):
    assert bound % 2 == 1, "bound must be odd"
    side = bound // 2
    dim = min(signal.shape)
    dimB = max(signal.shape)
    padded = np.concatenate(
        [np.zeros([dim, side]), signal, np.zeros([dim, side])], axis=1
    )
    mat = [np.arange(i, dimB + i, stride) for i in range(bound)]
    return padded[:, mat]


def GetGradientMulti3(linepoints, span):
    padded = np.concatenate(
        [np.zeros([span, 3]), linepoints, np.zeros([span, 3])]
    )
    kernel = np.concatenate(
        [-np.arange(1, span + 1, 1), np.array([0]), np.arange(span, 0, -1)]
    )
    derivative = []
    for i in range(len(linepoints)):
        derivative.append(np.matmul(kernel, padded[i : i + span * 2 + 1]))
    return derivative


def Getdirdif3(linepoints, span):
    padded = np.concatenate(
        [np.zeros([span, 3]), linepoints, np.zeros([span, 3])]
    )
    # kernel = np.concatenate([np.ones(span),[span*-2],np.ones(span)])
    kernel = np.concatenate(
        [
            np.arange(1, span + 1) * 0.2 + 0.8,
            [-np.sum(np.arange(1, span + 1) * 0.2 + 0.8) * 2],
            np.arange(span, 0, -1) * 0.2 + 0.8,
        ]
    )
    derivative = []
    for i in range(len(linepoints)):
        derivative.append(np.matmul(kernel, padded[i : i + span * 2 + 1]))
    return derivative


def findcurves(
    points,
    threshold=0.005,
    minimumpoints=16,
    sigma=1,
    useGradient=False,
    usemul=True,
    mulspan1=2,
    mulspan2=2,
):
    # Debug()
    if sigma > 1:
        linepointsFiltered = Filter3dline(points, sigma)
    else:
        linepointsFiltered = bilateral3dNew(points.T, mulspan1, sigma)
    if useGradient:
        derivative1 = GetGradient(linepointsFiltered)
        derivative2 = GetGradient(derivative1)
    elif usemul:
        derivative1 = GetGradientMulti3(linepointsFiltered, mulspan1)
        derivative2 = GetGradientMulti3(derivative1, mulspan2)
    else:
        derivative2 = Getdirdif3(linepointsFiltered, mulspan2)
    segments = []
    segmentbuffer = []
    segmentAccumilator = []

    for n, i in enumerate(linepointsFiltered.tolist()):
        if (
            np.linalg.norm(derivative2[n])
            / np.linalg.norm(linepointsFiltered[n])
        ) > threshold:
            # if ((np.linalg.norm(derivative2[n])/np.linalg.norm(linepointsFiltered[n])) > threshold) | len(segmentbuffer) > 30:
            segments.append(i)
            if len(segmentbuffer) > minimumpoints:
                segmentAccumilator.append(segmentbuffer)
            segmentbuffer = []
        else:
            segmentbuffer.append(i)
    return segmentAccumilator


def groupsimilar(curves, distance=0.05, angle=20):
    PrinicpalVectors = []
    startpoint = []
    endpoint = []
    newgroups = []
    angleRAD = np.deg2rad(angle)
    breakpoint = True
    for n, i in enumerate(curves):
        savecentroid = np.sum(i, 0) / len(i)
        A = i - savecentroid
        VT = np.linalg.eig(np.matmul(A.T, A))
        eigval = VT[0]
        sort = sorted(zip(eigval, VT[1].T.tolist()), reverse=True)
        _, VT = zip(*sort)
        PrinicpalVectors.append(VT)
        startpoint.append(i[0])
        endpoint.append(i[-1])

    newgroups.append(curves[0])
    for n, i in enumerate(curves[:-1]):
        pointdistance = np.linalg.norm(
            np.subtract(endpoint[n], startpoint[n + 1])
        )
        pointangle = Corners.angle_b_vectors(
            PrinicpalVectors[n], PrinicpalVectors[n + 1]
        )
        if (pointdistance < distance) and (
            pointangle < angleRAD or pointangle > np.pi - angleRAD
        ):
            newgroups[-1] = newgroups[-1] + curves[n + 1]
        else:
            newgroups.append(curves[n + 1])
    return newgroups


def Beamplaneacumulate(
    BeamDF,
    threshold=0.05,
    minimumpoints=16,
    distance=0.1,
    angle=20,
    sigma=1,
    usemul=False,
    useGradient=False,
    mulspan1=2,
    mulspan2=3,
):
    BeamAcumulator = []
    lasersequence = [15, 13, 11, 9, 7, 5, 3, 1, 14, 12, 10, 8, 6, 4, 2, 0]
    for i in lasersequence:
        linepointsDF = BeamDF[(BeamDF["laser_id"] == (i))]
        # linepoints=linepointsDF.sort_values(by=['azimuth']).iloc[:,1:4].values
        linepoints = linepointsDF[["X", "Y", "Z"]].values
        # segmentAccumilator = findcurves(linepoints,threshold = 0.001,minimumpoints = 12)
        # segmentAccumilator = findcurves(linepoints,threshold = 0.05,minimumpoints = 16,
        #                              sigma = 1,usemul = True,mulspan1 = 3, mulspan2 = 3 )
        segmentAccumilator = findcurves(
            linepoints,
            threshold=threshold,
            minimumpoints=minimumpoints,
            sigma=sigma,
            usemul=usemul,
            useGradient=useGradient,
            mulspan1=mulspan1,
            mulspan2=mulspan2,
        )
        if distance > 0.00001:
            segmentAccumilator = groupsimilar(
                segmentAccumilator, distance=distance, angle=angle
            )
        BeamAcumulator.append(segmentAccumilator)
    return BeamAcumulator


def SmoothBeams(BeamDF, mulspan1=9, sigma=0.04):
    BeamAcumulator = []
    lasersequence = [15, 13, 11, 9, 7, 5, 3, 1, 14, 12, 10, 8, 6, 4, 2, 0]
    for i in lasersequence:
        linepointsDF = BeamDF[(BeamDF["laser_id"] == (i))]
        linepoints = linepointsDF[["X", "Y", "Z"]].values
        segmentAccumilator = bilateral3dNew(linepoints.T, mulspan1, sigma)
        BeamAcumulator.append(segmentAccumilator)
    return BeamAcumulator


def findplaneproposal2(initial, centroids, curves, PVs):
    curvenumber = initial
    closestpointsid = findclosest(np.array(centroids[curvenumber]), centroids)
    n = 1
    success = False
    options = []
    votes = []
    while n < 10:
        closestpoints = curves[closestpointsid[0]] + curves[closestpointsid[n]]
        val, vectorspace, vectorcentroid = getPrincipalComponents(
            closestpoints
        )
        normalvector = homovector(vectorspace[2])
        pointangle = Corners.angle_b_vectors(
            PVs[curvenumber][0], PVs[closestpointsid[n]][0]
        )
        distance = np.linalg.norm(
            centroids[closestpointsid[0]] - centroids[closestpointsid[n]]
        )
        if distance > (0.4 * np.linalg.norm(centroids[closestpointsid[0]])):
            break

        if (
            (val[0] / val[1] < 80)
            & (pointangle < 0.34906 or pointangle > np.pi - 0.34906)
            & (val[1] / val[2] > 200)
        ):
            found = False
            for nn, i in enumerate(options):
                angg = Corners.angle_b_vectors(normalvector, i)
                if pointangle < 0.17 or pointangle > np.pi - 0.17:
                    votes[nn] += 1
                    found = True
            if found == False:
                options.append(normalvector.tolist())
                votes.append(1)
        n += 1

    if len(votes) > 0:
        sort = sorted(zip(votes, options), reverse=True)
        votes, options = zip(*sort)
        votes = list(votes)
        options = list(options)

    return options, votes


def cart2sph(x, y, z):
    hxy = np.hypot(x, y)
    r = np.hypot(hxy, z)
    el = np.arctan2(z, hxy)
    az = np.arctan2(y, x)
    return az, el, r


def sph2cart(az, el, r):
    rcos_theta = r * np.cos(el)
    x = rcos_theta * np.cos(az)
    y = rcos_theta * np.sin(az)
    z = r * np.sin(el)
    return x, y, z


def homovector(vector):
    assert len(vector) == 3, "vector must be dim 3"
    assert (
        (vector[0] != 0) | (vector[1] != 0) | (vector[2] != 0)
    ), "vector is trivial"
    s1 = np.sign(vector[0])
    s2 = np.sign(vector[1])
    s3 = np.sign(vector[2])

    if s1 != 0:
        vector = np.multiply(vector, s1)
    elif s2 != 0:
        vector = np.multiply(vector, s2)
    elif s3 != 0:
        vector = np.multiply(vector, s3)
    return vector


def similarize(vector, target):
    Nvector = np.array(vector)
    assert len(Nvector) == 3, "vector must be dim 3"
    angle = Corners.angle_b_vectors(Nvector, target)
    if angle > np.pi / 2:
        Nvector = -Nvector
    return Nvector


def preparedata(acu):
    AllCurves = [o for i in acu for o in i]
    Allcentroids = []
    AllPV = []

    for i in AllCurves:
        val, vec, cen = getPrincipalComponents(i)
        Allcentroids.append(cen)
        AllPV.append(vec)

    return AllCurves, Allcentroids, AllPV


def findplane():
    pass


def planefinderHough2(BoundDF):
    verbosity = True
    # acu = Beamplaneacumulate(BoundDF,threshold = 0.14, minimumpoints = 14, distance = 0.001, angle = 0.5, sigma = 5, usemul = False, mulspan1 = 2, mulspan2 = 8 ,useGradient = False)

    acu = Beamplaneacumulate(
        BoundDF,
        threshold=0.2,
        minimumpoints=12,
        distance=0.1,
        angle=5,
        sigma=0.08,
        usemul=False,
        mulspan1=21,
        mulspan2=21,
        useGradient=False,
    )
    AllCurves, Allcentroids, AllPV = preparedata(acu)
    worklines = np.array(AllCurves)
    workcentroids = np.array(Allcentroids)
    workPVs = np.array(AllPV)
    listoptions = []
    listvotes = []
    listlines = []
    listcenters = []
    for n, i in enumerate(worklines):
        options, votes = findplaneproposal2(
            n, workcentroids, worklines, workPVs
        )
        if len(votes) > 0:
            listlines.append(worklines[n])
            listoptions.append(options[0])
            listvotes.append(votes)
            listcenters.append(workcentroids[n])

    maxlen = max([len(i) for i in listvotes])
    buff = [list(i) + np.zeros(maxlen - len(i)).tolist() for i in listvotes]
    probs = scipy.special.softmax(buff, 1)
    goodcurves = np.max(probs, 1) > 0.6

    planesslot = np.array(listlines)[goodcurves].tolist()
    goodnormals = np.array(listoptions)[goodcurves].tolist()
    goodcenters = np.array(listcenters)[goodcurves].tolist()

    planes = []
    centers = []

    sphericals = np.array([cart2sph(i[0], i[1], i[2]) for i in goodnormals])

    db2 = DBSCAN(eps=0.1, min_samples=5).fit(sphericals[:, 0:2])
    for i in np.unique(db2.labels_):
        if i == -1:
            continue
        pp = np.array(planesslot)[db2.labels_ == i].tolist()
        centers.append(np.array(goodcenters)[db2.labels_ == i])
        planes.append(pp)
    finalplanes = []
    finalcenters = []

    for i in range(len(planes)):
        points = [o for i2 in planes[i] for o in i2]
        left = len(points)
        buffer = np.array(points)
        while left > len(points) * 0.05:

            plane, mask, _ = findplane(buffer, 0.06)
            if len(plane) < 200:
                break
            finalplanes.append(plane)
            left = len(buffer[~mask])
            buffer = buffer[~mask]

    return finalplanes
    # return finalplanes,planes,planesslot,goodnormals,goodcenters


def planefinderHough2diag(BoundDF):
    verbosity = True
    # acu = Beamplaneacumulate(BoundDF,threshold = 0.14, minimumpoints = 14, distance = 0.001, angle = 0.5, sigma = 5, usemul = False, mulspan1 = 2, mulspan2 = 8 ,useGradient = False)

    acu = Beamplaneacumulate(
        BoundDF,
        threshold=0.2,
        minimumpoints=12,
        distance=0.000002,
        angle=1,
        sigma=0.08,
        usemul=False,
        mulspan1=21,
        mulspan2=21,
        useGradient=False,
    )
    AllCurves, Allcentroids, AllPV = preparedata(acu)
    worklines = np.array(AllCurves)
    workcentroids = np.array(Allcentroids)
    workPVs = np.array(AllPV)
    listoptions = []
    listvotes = []
    listlines = []
    listcenters = []
    for n, i in enumerate(worklines):
        options, votes = findplaneproposal2(
            n, workcentroids, worklines, workPVs
        )
        if len(votes) > 0:
            listlines.append(worklines[n])
            listoptions.append(options[0])
            listvotes.append(votes)
            listcenters.append(workcentroids[n])

    maxlen = max([len(i) for i in listvotes])
    buff = [list(i) + np.zeros(maxlen - len(i)).tolist() for i in listvotes]
    probs = scipy.special.softmax(buff, 1)
    goodcurves = np.max(probs, 1) > 0.6

    planesslot = np.array(listlines)[goodcurves].tolist()
    goodnormals = np.array(listoptions)[goodcurves].tolist()
    goodcenters = np.array(listcenters)[goodcurves].tolist()

    planes = []
    centers = []

    sphericals = np.array([cart2sph(i[0], i[1], i[2]) for i in goodnormals])

    db2 = DBSCAN(eps=0.1, min_samples=5).fit(sphericals[:, 0:2])
    for i in np.unique(db2.labels_):
        if i == -1:
            continue
        pp = np.array(planesslot)[db2.labels_ == i].tolist()
        centers.append(np.array(goodcenters)[db2.labels_ == i])
        planes.append(pp)
    finalplanes = []
    finalcenters = []

    for i in range(len(planes)):
        points = [o for i2 in planes[i] for o in i2]
        left = len(points)
        buffer = np.array(points)
        while left > len(points) * 0.05:

            plane, mask, _ = findplane(buffer, 0.06)
            if len(plane) < 200:
                break
            finalplanes.append(plane)
            left = len(buffer[~mask])
            buffer = buffer[~mask]

    return finalplanes, planes, planesslot, goodnormals, goodcenters


def makesphere(centroid=[0, 0, 0], radius=1, dense=90):
    n = np.arange(0, 360, int(360 / dense))
    n = np.deg2rad(n)
    x, y = np.meshgrid(n, n)
    x = x.flatten()
    y = y.flatten()
    sphere = np.vstack(
        [
            centroid[0] + np.sin(x) * np.cos(y) * radius,
            centroid[1] + np.sin(x) * np.sin(y) * radius,
            centroid[2] + np.cos(x) * radius,
        ]
    ).T
    return sphere


"""    
class ArucoCam:
    def __init__(self,hasIntrinsic=True, file=None):
        if hasIntrinsic:
            with np.load(file+'.npz') as L:
                self.camera_matrix, self.dist_coefs, self.new_camera_matrix, self.seedH, self.seedW =\
                [L[i] for i in ('camera_matrix','dist_coefs','new_camera_matrix','seedH','seedW')]
            
    def findCharucos(frame,arucodict):
        arucodict1=aruco.getPredefinedDictionary(aruco.DICT_4X4_250)
        arucodict2=aruco.getPredefinedDictionary(aruco.DICT_6X6_250)
        arucodict3=aruco.getPredefinedDictionary(aruco.DICT_5X5_250)
        markersize = 0.0848
        squaresize = 0.1062
        #squaresize = 0.08482
        charboard1=aruco.CharucoBoard_create(9,7,squaresize,markersize,arucodict1) 
        charboard2=aruco.CharucoBoard_create(9,7,squaresize,markersize,arucodict2) 
        charboard3=aruco.CharucoBoard_create(9,7,squaresize,markersize,arucodict3) 
        paramiters=cv2.aruco.DetectorParameters_create()
        paramiters.cornerRefinementMethod = cv2.aruco.CORNER_REFINE_SUBPIX
        paramiters.adaptiveThreshWinSizeMax = 33
        paramiters.adaptiveThreshWinSizeStep = 16

        rot1=[]
        listids1=[]
        tra1=[]
        rot2=[]
        listids2=[]
        tra2=[]
        rot3=[]
        listids3=[]
        tra3=[]

        findingaruco1,ids1,reg1=aruco.detectMarkers(frame,arucodict1,parameters =paramiters)
        if len(findingaruco1)>0:
            findingaruco2,ids2,reg2=aruco.detectMarkers(frame,arucodict2,parameters =paramiters)
            if len(findingaruco2)>0:
                findingaruco3,ids3,reg3=aruco.detectMarkers(frame,arucodict3,parameters =paramiters)
                if len(findingaruco3)>0:            
                    rvec1,tvec1,objectpoints1=aruco.estimatePoseSingleMarkers(findingaruco1,markersize,camera_matrix,dist_coefs)
                    rvec2,tvec2,objectpoints2=aruco.estimatePoseSingleMarkers(findingaruco2,markersize,camera_matrix,dist_coefs)
                    rvec3,tvec3,objectpoints3=aruco.estimatePoseSingleMarkers(findingaruco3,markersize,camera_matrix,dist_coefs)
                    rot1.append(rvec1[ids1.argsort(0)].squeeze())
                    tra1.append(tvec1[ids1.argsort(0)].squeeze())
                    listids1.append(ids1[ids1.argsort(0)].squeeze())
                    rot2.append(rvec2[ids2.argsort(0)].squeeze())
                    tra2.append(tvec2[ids2.argsort(0)].squeeze())
                    listids2.append(ids2[ids2.argsort(0)].squeeze())
                    rot3.append(rvec3[ids3.argsort(0)].squeeze())
                    tra3.append(tvec3[ids3.argsort(0)].squeeze())
                    listids3.append(ids3[ids3.argsort(0)].squeeze())
        IntCor1 = cv2.aruco.interpolateCornersCharuco(findingaruco1,ids1,frame,charboard1,cameraMatrix =camera_matrix,distCoeffs =dist_coefs)
        _,Brvec1,Btvec1=cv2.aruco.estimatePoseCharucoBoard(IntCor1[1], IntCor1[2], charboard1, camera_matrix, dist_coefs)

        IntCor2 = cv2.aruco.interpolateCornersCharuco(findingaruco2,ids2,frame,charboard2,cameraMatrix =camera_matrix,distCoeffs =dist_coefs)
        _,Brvec2,Btvec2=cv2.aruco.estimatePoseCharucoBoard(IntCor2[1], IntCor2[2], charboard2, camera_matrix, dist_coefs)

        IntCor3 = cv2.aruco.interpolateCornersCharuco(findingaruco3,ids3,frame,charboard3,cameraMatrix =camera_matrix,distCoeffs =dist_coefs)
        _,Brvec3,Btvec3=cv2.aruco.estimatePoseCharucoBoard(IntCor3[1], IntCor3[2], charboard3, camera_matrix, dist_coefs)

        return Brvec1,Btvec1,Brvec2,Btvec2,Brvec3,Btvec3
    
    def findMarkersCharuco(image = None,file = None , chardict = 4, , equalize = 0):
        assert 4<=chardict<=6,'Chardict must be between 4 and 6'
        assert (image is not None) != (file is not None),'Need input image'
        
        allCorners  = [] # 3d point in real world space
        allIds  = []
        allDirs  = []
        
        if chardict ==4:
            self.arucodict = aruco.getPredefinedDictionary(aruco.DICT_4X4_250)
        elif chardict ==6:
            self.arucodict = aruco.getPredefinedDictionary(aruco.DICT_6X6_250)
        elif chardict ==5:
            self.arucodict = aruco.getPredefinedDictionary(aruco.DICT_5X5_250)
        
        paramiters=cv2.aruco.DetectorParameters_create()
        paramiters.cornerRefinementMethod = cv2.aruco.CORNER_REFINE_NONE
        paramiters.adaptiveThreshWinSizeMax = 103
        paramiters.adaptiveThreshWinSizeStep = 82
        
        if image is not None:
            imgc = image
            
        if image is not None:
            imgc = cv2.imread(fname)
            
        imgc = cv2.resize(imgc,(seedW,seedH))
        
        if equalize != 0:
            imgc = cv2.min(imgc,150)
            imgc = cv2.equalizeHist(imgc)
            
        corners, ids, rejectedImgPoints = cv2.aruco.detectMarkers(imgc, arucodict,parameters =paramiters)
        if hasIntrinsic:
            res2 = cv2.aruco.interpolateCornersCharuco(findingaruco1,ids1,frame,arucodict,cameraMatrix =self.camera_matrix,distCoeffs =self.dist_coefs)
        else:
            res2 = cv2.aruco.interpolateCornersCharuco(corners,ids,imgc,charboard1)
        self.corners = corners  
        self.ids = ids
        self.charcorners = res2[1]
        self.charids = res2[2]
        return corners,ids,res2
    
    def getcharucoPose():
    _,Brvec1,Btvec1=cv2.aruco.estimatePoseCharucoBoard(self.charcorners, self.charids, self.arucodict, self.camera_matrix, self.dist_coefs)
        rvec1,tvec1,objectpoints1=aruco.estimatePoseSingleMarkers(findingaruco1,markersize,camera_matrix,dist_coefs) 
"""



try:
    import open3d as o3d
    use_headless = False
except ImportError:
    use_headless = True
import numpy as np


def convertcloud(points):
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(points)
    return pcd


def downsample(point_cloud, leaf_size=0.005, return_idx=False):
    if return_idx:
        pcd = o3d.geometry.PointCloud()
        pcd.points = o3d.utility.Vector3dVector(point_cloud)
        _, rest, _ = pcd.voxel_down_sample_and_trace(
            voxel_size=leaf_size,
            min_bound=np.array([-10, -10, -10]),
            max_bound=np.array([10, 10, 10]),
        )
        return rest[rest != -1]
    else:
        return o3d.voxel_down_sample(point_cloud, leaf_size)


# def RHTPlanes(points)
load = False
if load == False:
    restart = False
    if restart == True:
        workpoints = pointsDF
        found = []

    maxtheta = np.pi
    mintheta = -np.pi
    maxp = 1
    minp = -1

    sizeofac = 6
    acumulator = np.zeros([sizeofac, sizeofac, sizeofac])
    threshhold = 40
    acumulatorchoices = np.zeros(
        [sizeofac, sizeofac, sizeofac, (threshhold + 1) * 3]
    )
    xstep = (maxtheta - mintheta) / sizeofac
    ystep = xstep
    zstep = (maxp - minp) / sizeofac
    mindist = 0.08
    distance = 0.15
    maxiterations = 400000
    iteration = 0
    while len(workpoints) > 2:
        iteration += 1
        choices = np.random.choice(len(workpoints), 3)
        points = workpoints[choices]

        if iteration % 50000 == 0:
            print(iteration)

        if iteration > maxiterations:
            break

        if (
            (np.linalg.norm(points[0] - points[1]) < distance)
            & (np.linalg.norm(points[2] - points[1]) < distance)
            & (np.linalg.norm(points[2] - points[0]) < distance)
            & (np.linalg.norm(points[0] - points[1]) > mindist)
            & (np.linalg.norm(points[2] - points[1]) > mindist)
            & (np.linalg.norm(points[2] - points[0]) > mindist)
        ):
            P = np.dot(
                np.cross(points[2] - points[1], points[0] - points[1]),
                points[0],
            )
            x0 = [0, 0]
            arges = scipy.optimize.least_squares(
                Corners.hugh3p_inter,
                x0,
                args=(points[0], points[1], points[2]),
                ftol=1e-6,
                xtol=1e-6,
                gtol=1e-6,
                max_nfev=10,
                bounds=[[-np.pi, -np.pi], [np.pi, np.pi]],
            )
            if arges.success == True:
                theta = int((arges.x[0] + maxtheta) / xstep)
                phi = int((arges.x[1] + maxtheta) / ystep)
                # p=int((hughtransform_1p(points[0],arges.x[0],arges.x[1])+maxp)/(zstep))
                p = int((P + maxp) / (zstep))

                acumulatorchoices[
                    theta, phi, p, int(acumulator[theta, phi, p] * 3)
                ] = choices[0]
                acumulatorchoices[
                    theta, phi, p, int(acumulator[theta, phi, p] * 3 + 1)
                ] = choices[1]
                acumulatorchoices[
                    theta, phi, p, int(acumulator[theta, phi, p] * 3 + 2)
                ] = choices[2]

                acumulator[theta, phi, p] += 1

                if acumulator[theta, phi, p] > threshhold:
                    print("planefound")
                    choicepoints = np.int32(
                        acumulatorchoices[theta, phi, p, :]
                    )
                    foundpoints = workpoints[
                        np.int32(acumulatorchoices[theta, phi, p, :])
                    ]
                    found.append([theta, phi, p, foundpoints])
                    # planepoint
                    workpoints = np.delete(workpoints, choicepoints, 0)
                    acumulator = np.zeros([sizeofac, sizeofac, sizeofac])
                    acumulatorchoices = np.zeros(
                        [sizeofac, sizeofac, sizeofac, (threshhold + 1) * 3]
                    )
    print("points left", workpoints.shape)
    print("planes found", len(found))


if load == False:
    mypoints = []
    my_list = []
    my_points = []
    for i in range(len(found)):
        my_list.append(found[i][0:3])
        my_points.append(found[i][3])

    myarr = np.array(my_list)
    uniques = np.unique(myarr, axis=0)
    print(uniques)
    for o in range(len(uniques)):
        mypoints.append([])

    for i in range(len(myarr)):
        for o in range(len(uniques)):
            if np.array_equal(myarr[i], uniques[o]):
                if len(mypoints[o]) == 0:
                    mypoints[o].append(my_points[i])
                else:
                    mypoints[o][0] = np.concatenate(
                        [mypoints[o][0], my_points[i]], 0
                    )
    totalpoints = 0
    for o in range(len(uniques)):
        totalpoints += len(mypoints[o][0])
        print(len(mypoints[o][0]))

    planemean = totalpoints / len(uniques)
    posibleplanes = []
    for o in range(len(uniques)):
        if len(mypoints[o][0]) > planemean:
            posibleplanes.append(mypoints[o][0])

if load == False:
    if len(uniques) <= 20:
        vis = open3d.Visualizer()
        vis.create_window()
        for i in range(len(uniques)):
            colNORM = i / len(uniques)
            color = cm.jet(colNORM)[:3]
            pointcloud1 = Corners.convertcloud(mypoints[i][0], Path)
            pointcloud1.paint_uniform_color(color)
            vis.add_geometry(pointcloud1)
        vis.run()
        vis.destroy_window()
    else:
        vis = open3d.Visualizer()
        vis.create_window()
        for i in range(len(posibleplanes)):
            colNORM = i / len(posibleplanes)
            color = cm.jet(colNORM)[:3]
            pointcloud1 = Corners.convertcloud(posibleplanes[i], Path)
            pointcloud1.paint_uniform_color(color)
            vis.add_geometry(pointcloud1)
        vis.run()
        vis.destroy_window()

if load == False:
    vis = open3d.Visualizer()
    vis.create_window()
    layers = [0, 1]
    for i in layers:
        colNORM = i / 5
        color = cm.jet(colNORM)[:3]
        pointcloud1 = Corners.convertcloud(mypoints[i][0], Path)
        pointcloud1.paint_uniform_color(color)
        vis.add_geometry(pointcloud1)
    vis.run()
    vis.destroy_window()


# OLD PLANE DETECTION
planenormals = []
planecentroid = []
planes = []
values = []

vectorspace, vectorcentroid, vectorspacepoints = Plane.FindPlanessparse(
    BoundDF
)
for i in vectorspacepoints:
    nfpp = np.array([o for o in i])
    orig = len(nfpp)
    while len(nfpp) / orig > 0.3:
        out = Corners.FitPlaneNormalValues(nfpp, thresh=0.4)
        nfpp = out[3]
        if len(out[2]) > 0:
            planecentroid.append(out[4])
            planenormals.append(out[0])
            planes.append(out[2])
            values.append(out[1])
        else:
            break
len(planes), len(vectorspacepoints)


finalplanes = []
badplanepoints = []
bufferplanepoints = planepoints
for i in range(len(planenormals)):
    planepointdist = Plane.findPlanePointswithcentroid(
        planenormals[i], planecentroid[i], bufferplanepoints
    )
    cond = np.array(planepointdist) < 0.06
    goodplanepoints = bufferplanepoints[cond]
    badplanepoints = +bufferplanepoints[cond]
    if len(goodplanepoints) > 100:
        bufferplanepoints = bufferplanepoints[~cond]
        finalplanes.append(goodplanepoints)
Corners.open3dpaint(finalplanes, axis=True)



import numpy as np


def point_dis(a, b=None):
    if b is not None:
        c = a - b
    else:
        c = a
    anorm = np.sqrt(np.sum(np.multiply(c, c)))
    return anorm

def angle_b_vectors(a, b):
    value = np.sum(np.multiply(a, b)) / (np.linalg.norm(a) * np.linalg.norm(b))
    if (value < -1) | (value > 1):
        value = np.sign(value)
    angle = np.arccos(value)
    return angle


def getPrincipalVectors(
    A,
):  # get pricipal vectors and values of a matrix centered around (0,0,0)
    VT = np.linalg.eig(np.matmul(A.T, A))
    sort = sorted(zip(VT[0], VT[1].T.tolist()), reverse=True)
    Values, Vectors = zip(*sort)
    return Vectors, Values


def Dpoint2Plane(
    point, planecoefs
):  # get minimum distance from a point to a plane
    return (
        planecoefs[0] * point[0]
        + planecoefs[1] * point[1]
        + planecoefs[2] * point[2]
        + planecoefs[3]
    ) / np.linalg.norm(planecoefs[0:3])


def DistPoint2Line(
    point, linepoint1, linepoint2=np.array([0, 0, 0])
):  # get minimum destance from a point to a line
    return np.linalg.norm(
        np.cross((point - linepoint2), (point - linepoint1))
    ) / np.linalg.norm(linepoint1 - linepoint2)


def rotation_matrix_from_vectors(vec1, vec2):
    if all(np.abs(vec1) == np.abs(vec2)):
        return np.eye(3)
    a, b = (vec1 / np.linalg.norm(vec1)).reshape(3), (
        vec2 / np.linalg.norm(vec2)
    ).reshape(3)
    v = np.cross(a, b)
    c = np.dot(a, b)
    s = np.linalg.norm(v)
    kmat = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
    rotation_matrix = np.eye(3) + kmat + kmat.dot(kmat) * ((1 - c) / (s ** 2))
    return rotation_matrix








import time
import numpy as np
import pandas as pd
from Plane import tic, toc
from pypacker import ppcap
from pypacker.layer12 import ethernet
from pypacker.layer3 import ip
from pypacker.layer4 import tcp


class VVdataloader:
    angles = [-15, 1, -13, 3, -11, 5, -9, 7, -7, 9, -5, 11, -3, 13, -1, 15]
    verticalcorrection = [
        0.0112,
        -0.0007,
        0.0097,
        -0.0022,
        0.0081,
        -0.0037,
        0.0066,
        -0.0051,
        0.0051,
        -0.0066,
        0.0037,
        -0.0081,
        0.0022,
        -0.0097,
        0.007,
        -0.00112,
    ]

    def __init__(self, file):
        # self.capfile = savefile.load_savefile(open(file, 'rb'), verbose=True)
        self.capfile = ppcap.Reader(filename=file)
        # self.packets = [i.raw() for i in self.capfile.packets]
        self.packets = []
        self.packets = [buf for ts, buf in self.capfile]
        self.packets = [i for i in self.packets if i[42:44] == b"\xff\xee"]
        self.dualreturn = (
            int.from_bytes(self.packets[0][1246:1247], byteorder="big") == 57
        )
        self.packettimestamplist = self.listoftimestamps(self.packets)
        self.timing_offsets = self.make_table(self.dualreturn)
        self.blockazimuths = self.azimuthblocklist()
        self.flatazimuths = [n for i in self.blockazimuths for n in i]
        self.frames = self.findframes()
        # pd.set_option('float_format', '{:f}'.format)

    def getMultiFrame(self, start, NF):
        n = start
        for i in range(NF):
            n += i
            if i == 0:
                LIDARDFpoints = self.getdataFrame(n)
            else:
                LIDARDFpoints = pd.concat(
                    [LIDARDFpoints, self.getdataFrame(n)]
                )
        LIDARDFpoints = LIDARDFpoints.sort_values("Azimuth")
        LIDARDFpoints = LIDARDFpoints.reset_index()
        LIDARpoints = LIDARDFpoints[["X", "Y", "Z"]].values

        return LIDARDFpoints, LIDARpoints

    def getdataFrame(self, frame):
        start = [0, 0]
        oneshot = True
        if frame > 0:
            start = [self.frames[frame - 1][0], self.frames[frame - 1][1]]
        end = [self.frames[frame][0] + 1, self.frames[frame][1]]

        points = []
        for i in range(start[0], end[0]):
            for blocknumber in range(12):
                firstazimuth = self.flatazimuths[i * 12 + blocknumber]
                if firstazimuth != self.flatazimuths[i * 12 + blocknumber + 1]:
                    secondazimuth = self.flatazimuths[i * 12 + blocknumber + 1]
                else:
                    secondazimuth = self.flatazimuths[i * 12 + blocknumber + 2]
                if secondazimuth < firstazimuth:
                    AzimuthGap = secondazimuth - firstazimuth + 360
                else:
                    AzimuthGap = secondazimuth - firstazimuth

                index = 4
                bytepart = self.packets[i][
                    42 + blocknumber * 100 : 142 + blocknumber * 100
                ]
                for m in range(32):

                    if m < 16:
                        Precision_Azimuth = (
                            firstazimuth + (AzimuthGap * 2.304 * m) / 55.296
                        )
                    else:
                        Precision_Azimuth = firstazimuth + (
                            AzimuthGap * 2.304 * ((m - 16) + 55.296)
                        ) / (2 * 55.296)
                    if Precision_Azimuth >= 360:
                        Precision_Azimuth = Precision_Azimuth - 360
                    distance, reflectance = self.singledata(
                        bytepart[index : index + 3]
                    )
                    angle = self.angles[m % 16]
                    X, Y, Z = self.calcPOS(
                        distance, Precision_Azimuth, angle, m % 16
                    )
                    if self.dualreturn and blocknumber % 2 == 1:
                        if (
                            self.singledata(bytepart[index : index + 3])[0]
                            != self.singledata(lastblock[index : index + 3])[0]
                        ):
                            points.append(
                                [distance, reflectance, Precision_Azimuth]
                                + [X, Y, Z]
                                + [
                                    self.packettimestamplist[i]
                                    + self.timing_offsets[m][blocknumber]
                                ]
                                + [m % 16]
                                + [angle]
                            )
                    else:
                        points.append(
                            [distance, reflectance, Precision_Azimuth]
                            + [X, Y, Z]
                            + [
                                self.packettimestamplist[i]
                                + self.timing_offsets[m][blocknumber]
                            ]
                            + [m % 16]
                            + [self.angles[m % 16]]
                        )
                    index += 3
                lastblock = bytepart
        points = list(filter(lambda x: x[0] != 0, points))
        start = next(n for n, x in enumerate(points) if x[2] < 10)
        end = len(points) - next(
            n for n, x in enumerate(reversed(points)) if x[2] > 350
        )
        points = points[start:end]
        pointsDF = pd.DataFrame(
            points,
            columns=[
                "Distance",
                "Intensity",
                "Azimuth",
                "X",
                "Y",
                "Z",
                "TimeStamp",
                "laser_id",
                "Angle",
            ],
        )
        pointsDF = pointsDF.sort_values(by=["TimeStamp"])
        return pointsDF

    def calcPOS(self, dist, azi, ang, n):
        alfa = np.deg2rad(azi)
        omega = np.deg2rad(ang)
        X = dist * np.cos(omega) * np.sin(alfa)
        Y = dist * np.cos(omega) * np.cos(alfa)
        Z = dist * np.sin(omega) - self.verticalcorrection[n]
        return X, Y, Z

    def findframes(self):
        oneshot = True
        framelist = []
        for y in range(len(self.blockazimuths)):
            for x in range(12):
                if oneshot:
                    oneshot = False
                    continue
                if x == 0:
                    if (
                        self.blockazimuths[y - 1][11]
                        > self.blockazimuths[y][0]
                    ):
                        framelist.append([y, 0])
                else:
                    if self.blockazimuths[y][x - 1] > self.blockazimuths[y][x]:
                        framelist.append([y, x])
        return framelist

    def azimuthblocklist(self):
        # packetazimuths = [[self.azimuthcalc(values[44+i2:46+i2]) for i2 in range(0,1200,100)] for values in self.packets]
        packetazimuths = [
            [
                int.from_bytes(values[44:46], byteorder="little") / 100,
                int.from_bytes(values[144:146], byteorder="little") / 100,
                int.from_bytes(values[244:246], byteorder="little") / 100,
                int.from_bytes(values[344:346], byteorder="little") / 100,
                int.from_bytes(values[444:446], byteorder="little") / 100,
                int.from_bytes(values[544:546], byteorder="little") / 100,
                int.from_bytes(values[644:646], byteorder="little") / 100,
                int.from_bytes(values[744:746], byteorder="little") / 100,
                int.from_bytes(values[844:846], byteorder="little") / 100,
                int.from_bytes(values[944:946], byteorder="little") / 100,
                int.from_bytes(values[1044:1046], byteorder="little") / 100,
                int.from_bytes(values[1144:1146], byteorder="little") / 100,
            ]
            for values in self.packets
        ]
        return packetazimuths

    def listoftimestamps(self, packets):
        simplelist = [self.timestampcalc(x[1242:1246]) for x in packets]
        return simplelist

    def distancecalc(self, byte1):
        return int.from_bytes(byte1, byteorder="little") * 0.002

    def azimuthcalc(self, byte1):
        return int.from_bytes(byte1, byteorder="little") / 100

    def timestampcalc(self, byte1):
        return int.from_bytes(byte1, byteorder="little")

    def singledata(self, bytepart):
        return [self.distancecalc(bytepart[0:2]), bytepart[2]]

    def make_table(self, dual_mode):
        timing_offsets = [
            [0.0 for x in range(12)] for y in range(32)
        ]  # Init matrix
        # constants
        full_firing_cycle = 55.296  # μs
        single_firing = 2.304  # μs
        # compute timing offsets
        for x in range(12):
            for y in range(32):
                if dual_mode:
                    dataBlockIndex = (x - (x % 2)) + (y // 16)
                else:
                    dataBlockIndex = (x * 2) + (y // 16)
                dataPointIndex = y % 16
                timing_offsets[y][x] = (full_firing_cycle * dataBlockIndex) + (
                    single_firing * dataPointIndex
                )
        return timing_offsets



import os
import time
from collections import defaultdict
from datetime import datetime
from turtle import color
import matplotlib.pyplot as plt
import numpy as np

try:
    import pandas as pd

    no_pandas = False
except:
    no_pandas = True


def get_timestamp():
    current_datetime = datetime.now()
    return current_datetime.strftime("%d-%m-%y_%X").replace(":", "-")


class timer:
    def __init__(self) -> None:
        self.clock_time = time.perf_counter()

    def tic(self):
        self.clock_time = time.perf_counter()

    def toc(self):
        return time.perf_counter() - self.clock_time

    def ttoc(self):
        val = self.toc()
        self.tic()
        return val

    def ptoc(self, message=None):
        print(message, self.toc())

    def pttoc(self, message=None):
        print(message, self.ttoc())


class timed_counter:
    def __init__(self, enabled=True) -> None:
        if enabled:
            self.timer = timer()
            self.counter = 0
            self.stop_time = 0
            self.stop_count = 0
        self.enabled = enabled

    def start(self):
        if self.enabled:
            if self.counter == 0:
                self.timer.tic()

    def stop(self):
        if self.enabled:
            self.stop_time = self.timer.toc()
            self.stop_count = self.counter

    def count(self):
        if self.enabled:
            self.counter += 1

    def get_frequency(self):
        if self.enabled:
            if self.stop_time == 0:
                return self.counter / self.timer.toc()
            else:
                return self.stop_count / self.stop_time

    def reset(self):
        if self.enabled:
            self.timer.tic()
            self.counter = 0

    def disable(self):
        self.enabled = False


class benchmarker:
    def __init__(self, file="performance/base") -> None:
        self.enable = True
        self.step_timer = timer()
        self.global_timer = timer()
        self.global_dict = []
        self.step_dict = defaultdict(int)
        self.file = file
        self.folder = "/".join(file.split("/")[:-1])
        self.started = False

    def enable(self):
        self.enable = True

    def disable(self):
        self.enable = False

    def start(self):
        if self.enable:
            self.step_timer.tic()
            self.global_timer.tic()
            self.started = True

    def gstep(self):
        if self.enable:
            self.gstop()
            self.step_dict = defaultdict(int)
            self.start()

    def gstop(self):
        if self.enable:
            if self.started:
                if not "global" in self.step_dict.keys():
                    self.step_dict["global"] = self.global_timer.ttoc()
                self.global_dict.append(self.step_dict)
                self.started = False

    def step(self, topic=""):
        if self.enable:
            self.step_dict[topic] += self.step_timer.ttoc()

    def data_summary(self):
        if self.enable:
            self.gstop()

            self.series = defaultdict(dict)
            for n, i in enumerate(self.global_dict):
                for key in i.keys():
                    self.series[key][n] = i[key]
            means = {
                k: [np.mean(list(self.series[k].values()))]
                for k, v in self.series.items()
            }
            if not no_pandas:
                df = pd.DataFrame().from_dict(means)
                os.makedirs(self.folder, exist_ok=True)
                df.to_csv(self.file + "_summary.csv")
                plt.figure(figsize=(18, 6))
                mymap = plt.get_cmap('jet')
                plt.title(os.path.basename(self.file) + '_bar')
                plt.tight_layout()
                rescale = lambda y: (y - np.min(y)) / (np.max(y) - np.min(y))
                plt.bar(np.arange(len(df.values[0])),df.values[0],label=list(df.keys()), color=mymap(rescale(df.values[0])))
                plt.legend(list(df.keys()))
                plt.savefig(self.file + "_bar.png", dpi=200)


    def plot_data(self):
        if self.enable:
            self.data_summary()
            series = self.series
            plt.figure(figsize=(18, 6))
            plt.title(os.path.basename(self.file))
            for keys in series.keys():
                X = np.array(list(series[keys].keys()))
                Y = np.array(list(series[keys].values()))

                Q1 = np.percentile(Y, 25, interpolation="midpoint")
                Q3 = np.percentile(Y, 75, interpolation="midpoint")
                IQR = Q3 - Q1
                bool_idx = (Y < (IQR + 1.5 * Q3)) & (Y > (IQR - 1.5 * Q1))
                X = X[bool_idx]
                Y = Y[bool_idx]

                plt.plot(X, Y)
            plt.tight_layout()
            plt.legend(list(series.keys()))
            plt.savefig(self.file + ".png", dpi=200)


class keydefaultdict(defaultdict):
    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        else:
            ret = self[key] = self.default_factory(key)
            return ret


class g_benchmarker:
    def __init__(self) -> None:
        self.benchmarkers = {}
        self.enable = True
        today = datetime.now()
        self.time_string = today.strftime("%d:%m:%Y:%H:%M")

    def enable(self):
        self.enable = True
        for bench in self.benchmarkers.values():
            bench.enable()

    def disable(self):
        self.enable = False
        for bench in self.benchmarkers.values():
            bench.disable()

    def __getitem__(self, item):
        get_bench = self.benchmarkers.get(item, None)
        if get_bench is None:
            self.benchmarkers[item] = benchmarker(
                f"performance_{self.time_string}/{item}"
            )
        return self.benchmarkers[item]

    def save(self):
        if self.enable:
            for bench in self.benchmarkers.values():
                bench.plot_data()


bench_dict = g_benchmarker()

g_timer1 = timer()
g_timer2 = timer()
g_timer3 = timer()



import cv2
import numpy as np
from matplotlib import pyplot as plt


def imshow(image):
    plt.imshow(image)
    plt.show()


def grid_image(image, size):
    out_image = np.zeros((size[1], size[0], 3), dtype=np.uint8)
    if type(image) is list:
        if len(image) > 0:
            rows = int(np.ceil(np.sqrt(len(image))))
            colms = int(np.ceil(len(image) / rows))
            if len(image[0].shape) == 3:
                h, w, _ = image[0].shape
                newimages = [image[0]]
            else:
                h, w = image[0].shape
                newimages = [GRAY3Channels(image[0])]

            for i in image[1:]:
                if len(i.shape) == 3:
                    newimages.append(cv2.resize(i, (w, h)))
                else:
                    newimages.append(cv2.resize(GRAY3Channels(i), (w, h)))

            v_stacks = []
            for i in range(colms):
                h_stacks = []
                for j in range(rows):
                    if len(newimages) > (j + i * rows):
                        h_stacks.append(newimages[j + i * rows])
                    else:
                        blank_im = np.zeros_like(h_stacks[-1])
                        h, w = blank_im.shape[:2]

                        for j in range(3):
                            cv2.rectangle(
                                blank_im,
                                (int(w * 0.2 * j), int(h * 0.2 * j)),
                                (int(w * (1 - 0.2 * j)), int(h * (1 - 0.2 * j))),
                                (255, 255, 255),
                                max(int(min(h, w) * 0.05), 1),
                            )
                        h_stacks.append(blank_im)
                v_stacks.append(np.hstack(h_stacks))

            out_image = np.vstack(v_stacks)

    out_image = cv2.resize(out_image.copy(), size)
    return out_image


def GRAY3Channels(image, values=[1, 1, 1]):
    assert type(image) is np.ndarray
    if len(image.shape) == 2:
        newim = cv2.merge(
            [
                cv2.multiply(image, values[0], dtype=cv2.CV_8U),
                cv2.multiply(image, values[1], dtype=cv2.CV_8U),
                cv2.multiply(image, values[2], dtype=cv2.CV_8U),
            ]
        )
        newim = newim * 255 if np.max(newim) <= 1 else newim
        return newim
    else:
        return image


def ROIim(image, size=(1820, 980), coordinates=True):
    ratio = [1, 1]
    if size is not None:
        ratio = [image.shape[1] / size[0], image.shape[0] / size[1]]
        imageR = cv2.resize(image.copy(), size)
    x = cv2.selectROI(imageR)
    # cv2.waitKey(0)
    cv2.destroyAllWindows()
    x2 = [
        int(x[0] * ratio[0]),
        int(x[1] * ratio[1]),
        int(x[2] * ratio[0]),
        int(x[3] * ratio[1]),
    ]
    if coordinates:
        return x2
    else:
        return image.copy()[x2[1] : x2[1] + x2[3], x2[0] : x2[0] + x2[2]]
